#! /usr/bin/env python
# ==========================================================================
# Generates a lightcurve.
#
# Copyright (C) 2014-2019 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools
import cscripts
from cscripts import obsutils
from cscripts import ioutils
from cscripts import mputils


# ================ #
# cslightcrv class #
# ================ #
class cslightcrv(ctools.csobservation):
    """
    Generates a lightcurve

    The cslightcrv class generates a light curve for Imaging Air Cherenkov
    Telescope event data by performing a maximum likelihood fit using
    ctlike in a series of time bins. The time bins can be either
    specified in an ASCII file, as an interval divided into equally
    sized time bins, or can be taken from the Good Time Intervals of the
    observation(s).

    The format of the ASCII file is one row per time bin, each specifying
    the start of stop value of the bin, separated by a whitespace. The
    times are given in Modified Julian Days (MJD). 

    Examples:
            >>> lcrv = cslightcrv()
            >>> lcrv.run()
            >>> ... (querying for parameters) ...
            >>> fits = lcrv.lightcurve()
                Generates a light curve and retrieves the results in
                a FITS file.

            >>> lcrv = cslightcrv()
            >>> lcrv.execute()
            >>> ... (querying for parameters) ...
                Generates a light curve and saves results in a FITS file.

            >>> lcrv = cslightcrv(obs)
            >>> lcrv.execute()
            >>> ... (querying for parameters) ...
                Generates a light curve from the observations in an
                observation container and saves results in a FITS file.
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)

        # Initialise some members
        self._srcname      = ''
        self._tbins        = gammalib.GGti()
        self._onoff        = False
        self._stacked      = False
        self._fits         = gammalib.GFits()
        self._nthreads     = 0
        self._excl_reg_map = None # Exclusion region map for on/off analysis

        # Return
        return

    # State methods por pickling
    def __getstate__(self):
        """
        Extend ctools.csobservation getstate method to include some members

        Returns
        -------
        state : dict
            Pickled instance
        """
        # Set pickled dictionary
        state = {'base'         : ctools.csobservation.__getstate__(self),
                 'srcname'      : self._srcname,
                 'tbins'        : self._tbins,
                 'stacked'      : self._stacked,
                 'onoff'        : self._onoff,
                 'fits'         : self._fits,
                 'nthreads'     : self._nthreads,
                 'excl_reg_map' : self._excl_reg_map}

        # Return pickled dictionary
        return state

    def __setstate__(self, state):
        """
        Extend ctools.csobservation setstate method to include some members

        Parameters
        ----------
        state : dict
            Pickled instance
        """
        ctools.csobservation.__setstate__(self, state['base'])
        self._srcname      = state['srcname']
        self._tbins        = state['tbins']
        self._stacked      = state['stacked']
        self._onoff        = state['onoff']
        self._fits         = state['fits']
        self._nthreads     = state['nthreads']
        self._excl_reg_map = state['excl_reg_map']

        # Return
        return

    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile
        """
        # Setup observations (require response and allow event list, don't
        # allow counts cube)
        self._setup_observations(self.obs(), True, True, False)

        # Set observation statistic
        self._set_obs_statistic(gammalib.toupper(self['statistic'].string()))

        # Set models if there are none in the container
        if self.obs().models().size() == 0:
            self.obs().models(self['inmodel'].filename())

        # Get time boundaries
        self._tbins = self._create_tbounds()

        # Set On/Off analysis flag and query relevant user parameters
        self._onoff = self._is_onoff()

        # Get source name
        self._srcname = self['srcname'].string()

        # Set stacked analysis flag and query relevant user parameters
        if not self._onoff:
            self._stacked = self._is_stacked()

        # Query the hidden parameters, just in case
        self['edisp'].boolean()
        self['calc_ulim'].boolean()
        self['calc_ts'].boolean()
        self['fix_bkg'].boolean()
        self['fix_srcs'].boolean()

        # Read ahead output parameters
        if self._read_ahead():
            self['outfile'].query()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Set number of processes for multiprocessing
        self._nthreads = mputils.nthreads(self)

        # Return
        return

    def _create_tbounds(self):
        """
        Creates light curve time bins

        Returns
        -------
        gti : `~gammalib.GGti`
            Light curve bins in form of Good Time Intervals
        """
        # Get MET time reference
        tref = gammalib.GTimeReference(self['mjdref'].real(),'s','TT','LOCAL')

        # Initialise Good Time Intervals
        gti = gammalib.GGti(tref)

        # Get algorithm to use for defining the time intervals. Possible
        # values are "FILE", "LIN" and "GTI". This is enforced at
        # parameter file level, hence no checking is needed.
        algorithm = self['tbinalg'].string()

        # If the algorithm is "FILE" then handle a FITS or an ASCII file for
        # the time bin definition
        if algorithm == 'FILE':

            # Get the filename
            filename = self['tbinfile'].filename()

            # If the file a FITS file then load GTIs directly
            if filename.is_fits():
                gti = gammalib.GGti(filename)

            # ... otherwise load file the ASCII file as CSV file and construct
            # the GTIs from the rows of the CSV file. It is expected that the
            # CSV file has two columns containing the "START" and "STOP"
            # values of the time bins. No header row is expected.
            else:
                csv = gammalib.GCsv(filename)
                for i in range(csv.nrows()):
                    tmin = gammalib.GTime()
                    tmax = gammalib.GTime()
                    tmin.mjd(csv.real(i,0))
                    tmax.mjd(csv.real(i,1))
                    gti.append(tmin,tmax)

        # If the algorithm is "LIN" then use a linear time binning, defined by
        # the "tmin", "tmax" and "tbins" user parameters
        elif algorithm == 'LIN':

            # Get start and stop time and number of time bins
            time_min = self['tmin'].time(tref)
            time_max = self['tmax'].time(tref)
            nbins    = self['tbins'].integer()

            # Compute time step in seconds and setup the GTIs
            time_step = (time_max - time_min) / float(nbins)
            for i in range(nbins):
                tmin = time_min +    i *time_step
                tmax = time_min + (i+1)*time_step
                gti.append(tmin,tmax)

        # If the algorithm is "GTI" then extract the GTIs from the observations
        # in the container and use them for the light curve time binning
        elif algorithm == 'GTI':

            # Append the GTIs of all observations
            for obs in self.obs():
                for i in range(obs.events().gti().size()):
                    gti.append(obs.events().gti().tstart(i),
                               obs.events().gti().tstop(i))

        # Return Good Time Intervals
        return gti

    def _get_free_par_names(self):
        """
        Return list of free parameter names

        Returns
        -------
        names : list of str
            List of free parameter names.
        """
        # Initialise empty list of free parameter names
        names = []

        # Collect list of free parameter names
        for par in self.obs().models()[self._srcname]:
            if par.is_free():
                names.append(par.name())

        # Return names
        return names

    def _adjust_model_pars(self):
        """
        Adjust model parameters dependent on user parameters
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Adjust model parameters')

        # Adjust model parameters dependent on input user parameters
        for model in self.obs().models():

            # Set TS flag for all models to false. The source of interest will
            # be set to true later
            model.tscalc(False)

            # Get the model classname to distinguish later sky from background
            # models
            classname = model.classname()

            # Log model name
            self._log_header3(gammalib.NORMAL, model.name())

            # If the model is the source of interest and the 'calc_ts' parameter
            # is true then enable the TS computation for the source
            if model.name() == self._srcname:
                if self['calc_ts'].boolean():
                    model.tscalc(True)

            # ... otherwise, if the model is not a sky model and the 'fix_bkg'
            # parameter is true or if the model is a sky model and the
            # 'fix_srcs' parameter is true then fix all parameters of the model
            elif ((self['fix_bkg'].boolean()  and classname != 'GModelSky') or
                  (self['fix_srcs'].boolean() and classname == 'GModelSky')):
                for par in model:
                    if par.is_free():
                        par.fix()
                        self._log_value(gammalib.NORMAL, par.name(), 'fixed')

        # Return
        return

    def _create_fits_table(self, results):
        """
        Creates FITS binary table containing light curve results

        Parameters
        ----------
        results : list of dict
            List of result dictionaries

        Returns
        -------
        table : `~gammalib.GFitsBinTable`
            FITS binary table containing light curve
        """
        # Determine number of rows in FITS table
        nrows = len(results)

        # Create FITS Table with extension "LIGHTCURVE"
        table = gammalib.GFitsBinTable(nrows)
        table.extname('LIGHTCURVE')

        # Append time columns
        ioutils.append_result_column(table, results, 'MJD',   'days', 'mjd')
        ioutils.append_result_column(table, results, 'e_MJD', 'days', 'e_mjd')

        # Append parameter columns
        ioutils.append_model_par_column(table,
                                        self.obs().models()[self._srcname],
                                        results)

        # Append Test Statistic column "TS"
        ioutils.append_result_column(table, results, 'TS', '', 'ts')

        # Append upper limit columns
        ioutils.append_result_column(table, results, 'DiffUpperLimit',
                                     'ph/cm2/s/MeV', 'ul_diff')
        ioutils.append_result_column(table, results, 'FluxUpperLimit',
                                     'ph/cm2/s', 'ul_flux')
        ioutils.append_result_column(table, results, 'EFluxUpperLimit',
                                     'erg/cm2/s', 'ul_eflux')

        # Return table
        return table

    def _compute_ulimit(self, obs):
        """
        Computes upper limit

        Parameters
        ----------
        obs : `~gammalib.GObservations`
            Observation container

        Returns
        -------
        ul_diff, ul_flux, ul_eflux : tuple of float
            Upper limits, -1.0 of not computed
        """
        # Initialise upper flux limit
        ul_diff  = -1.0
        ul_flux  = -1.0
        ul_eflux = -1.0

        # Perform computation only if requested
        if self['calc_ulim'].boolean():

            # Write header in logger
            self._log_header3(gammalib.EXPLICIT, 'Computing upper limit')

            # Create copy of observations
            cpy_obs = obs.copy()

            # Fix parameters of source of interest in copy of observations.
            # This assures that the original spectral parameters and position
            # are used in the upper limit search. The ctulimit tool makes sure
            # that the flux parameter is freed when needed.
            source = cpy_obs.models()[self._srcname]
            for par in source:
                if par.is_free():
                    par.fix()

            # Create instance of upper limit tool
            ulimit = ctools.ctulimit(cpy_obs)
            ulimit['srcname']   = self._srcname
            ulimit['eref']      = 1.0
            ulimit['emin']      = self['emin'].real()
            ulimit['emax']      = self['emax'].real()
            ulimit['sigma_min'] = 0.0
            ulimit['sigma_max'] = 3.0

            # Try to run upper limit tool and catch any exceptions
            try:
                ulimit.run()
                ul_diff  = ulimit.diff_ulimit()
                ul_flux  = ulimit.flux_ulimit()
                ul_eflux = ulimit.eflux_ulimit()
            except:
                self._log_string(gammalib.TERSE, 'Upper limit calculation failed.')
                ul_diff  = -1.0
                ul_flux  = -1.0
                ul_eflux = -1.0

        # Return upper limit tuple
        return ul_diff, ul_flux, ul_eflux

    def _timebin(self, i):
        """
        Run likelihood analysis in one time bin

        Parameters
        ----------
        i : int
            time bin number

        Returns
        -------
        result : dict
            Results of the likelihood analysis
        """
        # Get names of free parameters
        pars = self._get_free_par_names()

        # Get time boundaries
        tmin = self._tbins.tstart(i)
        tmax = self._tbins.tstop(i)

        # Write time bin into header
        self._log_header2(gammalib.TERSE, 'MJD %f - %f ' %
                          (tmin.mjd(), tmax.mjd()))

        # Compute time bin center and time width
        twidth = 0.5 * (tmax - tmin) # in seconds
        tmean  = tmin + twidth

        # Initialise result dictionary
        result = {'mjd': tmean.mjd(),
                  'e_mjd': twidth / gammalib.sec_in_day,
                  'ts': 0.0,
                  'ul_diff': 0.0,
                  'ul_flux': 0.0,
                  'ul_eflux': 0.0,
                  'pars': pars,
                  'values': {}}

        # Log information
        self._log_header3(gammalib.EXPLICIT, 'Selecting observations')

        # Select observations
        select = cscripts.csobsselect(self.obs())
        select['pntselect'] = 'CIRCLE'
        select['coordsys']  = 'GAL'
        select['glon']      =   0.0
        select['glat']      =   0.0
        select['rad']       = 180.0
        select['tmin']      = tmin
        select['tmax']      = tmax
        select.run()

        # Retrieve observations
        obs = select.obs()

        # If there are observations then select now events from them
        if obs.size() > 0:

            # Log information
            self._log_header3(gammalib.EXPLICIT, 'Selecting events')

            # Select events
            select = ctools.ctselect(obs)
            select['emin']  = self['emin'].real()
            select['emax']  = self['emax'].real()
            select['tmin']  = tmin
            select['tmax']  = tmax
            select['rad']   = 'UNDEFINED'
            select['ra']    = 'UNDEFINED'
            select['dec']   = 'UNDEFINED'
            select.run()

            # Retrieve observations
            obs = select.obs()

        # Continue only if there are observations
        if obs.size() > 0:

            # Deal with stacked and On/Off Observations
            if self._stacked or self._onoff:

                # If a stacked analysis is requested bin the events
                # and compute the stacked response functions and setup
                # an observation container with a single stacked observation.
                if self._stacked:
                    new_obs = obsutils.get_stacked_obs(self, obs, nthreads=1)

                # ... otherwise if On/Off analysis is requested generate
                # the On/Off observations and response
                elif self._onoff:
                    new_obs = obsutils.get_onoff_obs(self, obs, nthreads=1)

                # Extract models
                models = new_obs.models()

                # Fix background models if required
                if self['fix_bkg'].boolean():
                    for model in models:
                        if model.classname() != 'GModelSky':
                            for par in model:
                                par.fix()

                # Put back models
                new_obs.models(models)

                # Continue with new oberservation container
                obs = new_obs

            # Header
            self._log_header3(gammalib.EXPLICIT, 'Fitting the data')

            # Do maximum likelihood model fitting
            like = ctools.ctlike(obs)
            like['edisp']    = self['edisp'].boolean()
            like['nthreads'] = 1  # Avoids OpenMP conflict
            like.run()

            # Skip bin if no event was present
            if like.obs().logL() == 0.0:

                # Signal skipping of bin
                self._log_value(gammalib.TERSE, 'Warning',
                                'No event in this time bin, skip bin.')

                # Set all results to 0
                for par in pars:
                    result['values'][par]      = 0.0
                    result['values']['e_'+par] = 0.0

            # Otherwise fill in results dictionary
            else:
                # Retrieve model fitting results for source of interest
                source = like.obs().models()[self._srcname]

                # Extract parameter values
                for par in pars:
                    result['values'][par]      = source[par].value()
                    result['values']['e_'+par] = source[par].error()

                # Calculate upper limit (-1 if not computed)
                ul_diff, ul_flux, ul_eflux = self._compute_ulimit(obs)
                if ul_diff > 0.0:
                    result['ul_diff']  = ul_diff
                    result['ul_flux']  = ul_flux
                    result['ul_eflux'] = ul_eflux

                # Extract Test Statistic value
                if self['calc_ts'].boolean():
                    result['ts'] = source.ts()

                # Log results for this time bin
                self._log.header3('Results')
                pars = self._get_free_par_names()
                for par in pars:
                    value = source[par].value()
                    error = source[par].error()
                    unit  = source[par].unit()
                    self._log_value(gammalib.NORMAL, par,
                                    str(value)+' +/- '+str(error)+' '+unit)
                if ul_diff > 0.0:
                    self._log_value(gammalib.NORMAL, 'Upper flux limit',
                                    str(result['ul_diff'])+' ph/cm2/s/MeV')
                    self._log_value(gammalib.NORMAL, 'Upper flux limit',
                                    str(result['ul_flux'])+' ph/cm2/s')
                    self._log_value(gammalib.NORMAL, 'Upper flux limit',
                                    str(result['ul_eflux'])+' erg/cm2/s')
                if self['calc_ts'].boolean():
                    self._log_value(gammalib.NORMAL, 'Test Statistic', result['ts'])

        # Otherwise, if observations size is 0, signal bin is skipped and
        # fill results table with zeros
        else:
            self._log_value(gammalib.TERSE, 'Warning',
                            'No observations available in this time bin, '
                            'skip bin.')

            # Set all results to 0
            for par in pars:
                result['values'][par]        = 0.0
                result['values']['e_' + par] = 0.0

        # Return result dictionary
        return result


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Write observation into logger
        self._log_observations(gammalib.NORMAL, self.obs(), 'Observation')

        # Get time boundaries
        tmin = self._tbins.tstart(0)
        tmax = self._tbins.tstop(self._tbins.size()-1)

        # Select events
        select = ctools.ctselect(self.obs())
        select['emin'] = self['emin'].real()
        select['emax'] = self['emax'].real()
        select['tmin'] = tmin
        select['tmax'] = tmax
        select['rad']  = 'UNDEFINED'
        select['ra']   = 'UNDEFINED'
        select['dec']  = 'UNDEFINED'
        select.run()

        # Extract observations
        self.obs(select.obs().copy())

        # Write observation into logger
        self._log_header1(gammalib.TERSE,
                          gammalib.number('Selected observation',
                                          len(self.obs())))

        # Adjust model parameters dependent on user parameters
        self._adjust_model_pars()

        # Write header
        self._log_header1(gammalib.TERSE, 'Generate lightcurve')

        # If more than a single thread is requested then use multiprocessing
        if self._nthreads > 1:

            # Compute time bins
            args        = [(self, '_timebin', i)
                           for i in range(self._tbins.size())]
            poolresults = mputils.process(self._nthreads, mputils.mpfunc, args)

            # Construct results
            results = []
            for i in range(self._tbins.size()):
                results.append(poolresults[i][0])
                self._log_string(gammalib.TERSE, poolresults[i][1]['log'], False)

        # Otherwise loop over time bins and run time bin analysis
        else:
            results = []
            for i in range(self._tbins.size()):
                results.append(self._timebin(i))

        # Create FITS table from results
        table = self._create_fits_table(results)

        # Create FITS file and append FITS table to FITS file
        self._fits = gammalib.GFits()
        self._fits.append(table)

        # Optionally publish light curve
        if self['publish'].boolean():
            self.publish()

        # Return
        return

    def save(self):
        """
        Save light curve
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Save light curve')

        # Continue only if filename is valid
        if self['outfile'].is_valid():

            # Get light curve filename
            outfile = self['outfile'].filename()

            # Log file name
            self._log_value(gammalib.NORMAL, 'Light curve file', outfile.url())

            # Save spectrum
            self._fits.saveto(outfile, self._clobber())

        # Return
        return

    def publish(self, name=''):
        """
        Publish light curve

        Parameters
        ----------
        name : str, optional
            Name of light curve
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Publish light curve')

        # Continue only if light curve is valid
        if self._fits.contains('LIGHTCURVE'):

            # Set default name is user name is empty
            if not name:
                user_name = self._name()
            else:
                user_name = name

            # Log file name
            self._log_value(gammalib.NORMAL, 'Light curve name', user_name)

            # Publish light curve
            self._fits.publish('LIGHTCURVE', user_name)

        # Return
        return

    def lightcurve(self):
        """
        Return light curve FITS file

        Returns
        -------
        fits : `~gammalib.GFits()`
            FITS file containing light curve
        """
        # Return
        return self._fits

    def models(self, models):
        """
        Set model

        Parameters
        ----------
        models : `~gammalib.GModels()`
            Set model container
        """
        # Copy models
        self.obs().models(models)

        # Return
        return

    def exclusion_map(self, object=None):
        """
        Return and optionally set the exclusion regions map

        Parameters
        ----------
        object : `~gammalib.GSkyRegion` or `~gammalib.GSkyMap` or `~gammalib.GFilename`
            Exclusion regions object

        Returns
        -------
        region : `~gammalib.GSkyRegionMap`
            Exclusion regions map
        """
        # If a regions object is provided then set the regions ...
        if object is not None:
            self._excl_reg_map = gammalib.GSkyRegionMap(object)

        # Return
        return self._excl_reg_map


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = cslightcrv(sys.argv)

    # Execute application
    app.execute()



# ==========================================================================
# ctools Python module
#
# Copyright (C) 2015-2016 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
# import gammalib first
import gammalib

# Import sub modules
from .tools import *

# Remove swig register stuff
bad_entries = [entry for entry in list(locals())
               if entry.endswith('_swigregister')]
for entry in bad_entries:
    del locals()[entry]

# Clean-up
del tools
del gammalib
del bad_entries
del entry

# Add version
__version__ = '1.6.3'



#! /usr/bin/env python
# ==========================================================================
# Shows the distribution of significances in a given significance map.
#
# Copyright (C) 2018-2019 Andreas Specovius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# --------------------------------------------------------------------------
# What do I need for plotting:
#
# - significance map itself [mappath]
#   Provided as a file path guiding to a GSkyMap in FITS format
#   No default
#
# - included regions (for shrinking data space i.e. for cutting certain edges
#   or at a certain radius) [includedreg]
#   Provided as a file path to a FITS WCS map or DS9 regions file
#   Default should be None, which means all is allowed
#  
# - excluded regions [excludedreg]
#   Provided as a file path to a FITS WCS map or DS9 regions file
#   Default should be None, which means there are no excluded regions. In this
#   case no bkg will be plotted and fitted
#
# - binning parameters for the histogram: [nbins], [sign_min], [sign_max]
#   Provided as floats
#   Default should be -5 to 8 with 130 bins (131 edges)
#
# - title of the histogram [title]
#   Provided as string
#   Default should be an empty string
#
# - Path to plotted file [plotfile]
# ==========================================================================
import sys
import math
import gammalib
import cscripts
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()
try:
    import numpy as np
except ImportError:
    print('This script needs the "numpy" module')
    sys.exit()


# ======================= #
# Gaussian function class #
# ======================= #
class gaussian(gammalib.GPythonOptimizerFunction):

    # Constructor
    def __init__(self, x_vals, y_vals):
    
        # Call base class constructor
        gammalib.GPythonOptimizerFunction.__init__(self)
        
        # Set eval method
        self._set_eval(self.eval)

        # Set data
        self._x_vals = x_vals
        self._y_vals = y_vals

    # Methods
    def eval(self):
        """
        Evaluate function
        """
        # Recover parameters
        pars  = self._pars()
        norm  = pars[0].value()
        mean  = pars[1].value()
        sigma = pars[2].value()

        # Evaluate function values
        y = [norm * math.exp(-0.5*(x-mean)**2/(sigma**2)) for x in self._x_vals]

        # Compute weights (1/sqrt(y))
        weight = []
        for val in self._y_vals:
            if val > 0.0:
                weight.append(1.0/val)
            else:
                weight.append(0.0)
        #weight = [1.0/val if val > 0.0 else 0.0 for val in self._y_vals]

        # Compute Chi Square
        value = 0.0
        for i in range(len(self._x_vals)):
            arg    = self._y_vals[i] - y[i]
            value += arg * arg * weight[i]
        
        # Evaluate gradient and curvature
        sigma2 = sigma  * sigma
        sigma3 = sigma2 * sigma
        for i in range(len(self._x_vals)):

            # Evaluate function gradients
            dx     = self._x_vals[i] - mean
            dnorm  = y[i]         / norm   * pars[0].scale()
            dmean  = y[i] * dx    / sigma2 * pars[1].scale()
            dsigma = y[i] * dx**2 / sigma3 * pars[2].scale()

            # Setup gradient vector
            arg                 = (self._y_vals[i] - y[i]) * weight[i]
            self.gradient()[0] -= arg * dnorm
            self.gradient()[1] -= arg * dmean
            self.gradient()[2] -= arg * dsigma

            # Setup curvature matrix
            self.curvature()[0,0] +=  dnorm  * dnorm   * weight[i]
            self.curvature()[0,1] +=  dnorm  * dmean   * weight[i]
            self.curvature()[0,2] +=  dnorm  * dsigma  * weight[i]
            self.curvature()[1,0] +=  dmean  * dnorm   * weight[i]
            self.curvature()[1,1] +=  dmean  * dmean   * weight[i]
            self.curvature()[1,2] +=  dmean  * dsigma  * weight[i]
            self.curvature()[2,0] +=  dsigma * dnorm   * weight[i]
            self.curvature()[2,1] +=  dsigma * dmean   * weight[i]
            self.curvature()[2,2] +=  dsigma * dsigma  * weight[i]

        # Set value
        self._set_value(value)
        
        # Return
        return


# ================ #
# Helper functions #
# ================ #
def skymap_to_numpy_ndarray(skymap):
    """
    Return numpy.ndarray of pixel values of a skymap.

    Parameters
    ----------
    skymap : `~gammalib.GSkyMap`
        Sky map.

    Returns
    -------
    array : `~numpy.array`
        Numpy array
    """
    # Get map size
    npix  = skymap.npix()

    # Initialise out array
    array = np.empty(npix)

    # Loop over map pixels
    for ipix in range(npix):
        array[ipix] = skymap[ipix]

    # Return
    return array


def read_regions(fpath, init_map):
    """
    Read a regions WCS FITS or ds9 file and return a mask WCS map.

    Parameters
    ----------
    fpath : str
        Path to regions file
    init_map : `~gammalib.GSkyMap`
        Sky map representing fov of interest. Used for initialisation of
        regions map.

    Returns
    -------
    map : `~gammalib.GSkyMap`
        Sky map.
    """

    # Create starter map for user fov filled with zeros
    regions_map  = init_map.copy()
    regions_map *= 0.0

    # Take care about ds9 region files
    if fpath.lower().endswith('.reg'):

        # Read ds9 regions
        ds9_regions = gammalib.GSkyRegions(fpath)

        # Loop over regions
        for reg in ds9_regions:

            # Make map from region
            ds9_map = gammalib.GSkyRegionMap(reg)

            # Add ds9 region map to global map
            regions_map += ds9_map.map()

    # Take care about fits WCS files
    elif gammalib.GFilename(fpath).is_fits():

        # Read wcs regions map
        wcs_regions = gammalib.GSkyMap(fpath)

        # Add wcs regions map to global map
        regions_map += wcs_regions

    else:
        raise RuntimeError('Invalid regions file detected. Please provide ' +
                           'a valid ds9 or FITS WCS regions file.')

    # Return
    return regions_map


def linspace(minval, maxval, nbins):
    """
    Return a linearly spaced array
    
    Parameters
    ----------
    minval : float
        Minimum value
    maxval : float
        Maximum value
    nbins : int
        Number of bins

    Returns
    -------
    bins : list of float
        Linearly spaced array
    """
    # Compute bin width
    binwidth = (maxval - minval) / float(nbins-1)

    # Compute bins
    bins = [minval + binwidth*i for i in range(nbins)]

    # Return bins
    return bins


# ============================== #
# Plot significance distribution #
# ============================== #
def plot_significance_distribution(mappath, nbins, sigma_min, sigma_max,
                                   includedreg, excludedreg, title,
                                   plotfile):
    """
    Plot the significance distribution and return instance of pyplot
    figure and axis.

    Parameters
    ----------
    mappath : str
        Path to significance map
    nbins : int
        Number of bins in histogram
    sigma_min : float
        Lower limit of the x axis in the plot
    sigma_max : float
        Upper limit of the x axis in the plot
    includedreg : str
        Path to global included region file
    excludedreg : str
        Path to global excluded region file
    title : str
        Title of the plot
    plotfile : str
        Name of file for plotting
    """

    # Read significance map
    significance_map = gammalib.GSkyMap(mappath+'[SIGNIFICANCE]')


    # Read included regions
    inclusion_available = False
    if len(includedreg) > 0:
        regions_map         = read_regions(includedreg, init_map=significance_map)
        mask_include        = skymap_to_numpy_ndarray(regions_map).astype(bool)
        inclusion_available = True

    # Read excluded regions
    src_exclusion_available = False
    if len(excludedreg) > 0:
        regions_map             = read_regions(excludedreg, init_map=significance_map)
        mask_exclude            = skymap_to_numpy_ndarray(regions_map).astype(bool)
        src_exclusion_available = True

    # Convert significance map to flat array
    data = skymap_to_numpy_ndarray(significance_map)

    # Initialise dummy mask allowing all pixels
    mask = np.ones(significance_map.npix()).astype(bool)

    # Apply inclusion region spatial cuts
    if inclusion_available:
        mask &= mask_include
        data = data[mask]

    # Apply exclusion region spatial cuts
    if src_exclusion_available:
        data_without_sources = skymap_to_numpy_ndarray(significance_map)
        mask &= ~mask_exclude
        data_without_sources = data_without_sources[mask]

    # Compute bin edges, hence use nbins+1
    bin_edges = linspace(sigma_min, sigma_max, nbins+1)

    # Create figure
    fig = plt.figure()
    ax  = fig.gca()

    # Draw significance histogram for full FOV
    y, _, _ = ax.hist(data, bins=bin_edges, histtype='step', color='k',
                      label='significance')

    # Draw the default normal distribution. Scale to fit maximum of histogram
    yvals = [y.max() * math.exp(-0.5*(x-0.0)**2/(1.0**2)) for x in bin_edges]
    ax.plot(bin_edges, yvals, 'b-')
    msg = 'mean: $0$\nwidth: $1$'
    ax.text(0.98, 0.80, msg, ha='right', va='top',
            bbox=dict(edgecolor='blue', facecolor='white'),
            transform=ax.transAxes)

    # If exclusion was provided then also plot significances with excluded
    # source data
    if src_exclusion_available:

        # Draw histogram of significances for data with excluded sources
        y, _, _ = ax.hist(data_without_sources, bins=bin_edges, histtype='step',
                          color='k', alpha=0.5,
                          label='significance without exclusions')

        # Set initial Gaussian parameters
        y_max = float(y.max())
        par1  = gammalib.GOptimizerPar('Norm',  y_max)
        par2  = gammalib.GOptimizerPar('Mean',  0.0)
        par3  = gammalib.GOptimizerPar('Sigma', 1.0)
        pars  = gammalib.GOptimizerPars()
        pars.append(par1)
        pars.append(par2)
        pars.append(par3)

        # Set fit function
        x   = [0.5*(bin_edges[i]+bin_edges[i+1]) for i in range(nbins)]
        fct = gaussian(x, y)

        # Optimize function and compute errors
        opt = gammalib.GOptimizerLM()
        opt.optimize(fct, pars)
        opt.errors(fct, pars)

        # Recover parameters and errors
        norm    = pars[0].value()
        e_norm  = pars[0].error()
        mean    = pars[1].value()
        e_mean  = pars[1].error()
        sigma   = pars[2].value()
        e_sigma = pars[2].error()

        # Log fit result
        print('Fit of normal distribution to excluded source significance data '+
              'results in:  norm=%f+=%f  xmean=%f+-%f sigma=%f+-%f' %
              (norm, e_norm, mean, e_mean, sigma, e_sigma))

        # Draw text box
        msg = 'mean: $%.3f\pm%.3f$\nwidth: $%.3f\pm%.3f$' % \
              (mean, e_mean, sigma, e_sigma)
        ax.text(0.98, 0.95, msg, ha='right', va='top',
                bbox=dict(edgecolor='red', facecolor='white'),
                transform=ax.transAxes)

        # Plot the normal
        yvals = [norm * math.exp(-0.5*(x-mean)**2/(sigma**2)) for x in bin_edges]
        ax.plot(bin_edges, yvals, 'r-')

    # Configure the plot
    ax.set_ylim(0.5, ax.get_ylim()[1]*2.0)
    ax.set_xlim(sigma_min, sigma_max)
    ax.set_title(title)
    ax.set_xlabel('Significance')
    ax.set_ylabel('Entries')
    ax.grid()
    ax.set_yscale('log')

    # Show or save the plot
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ============================== #
# Show significance distribution #
# ============================== #
def show_significance_distribution():
    """
    Plot and show significance distribution
    """
    # Set usage string
    usage = 'show_significance_distribution.py [--nbins bins] ' +\
            '[--smin sigma_min] [--smax sigma_max] [--include regfile] ' +\
            '[--exclude regfile] [--title title] [-p plotfile] mapfile'

    # Set default options
    options = [{'option': '--nbins',    'value': '130'},
               {'option': '--smin',     'value':  '-5'},
               {'option': '--smax',     'value':   '8'},
               {'option': '--include',  'value':   ''},
               {'option': '--exclude',  'value':   ''},
               {'option': '--title',    'value':   ''},
               {'option': '-p',         'value':   ''}]

    # Get arguments and options from command line arguments
    args, options = cscripts.ioutils.get_args_options(options, usage)

    # Print help message
    if args[0]=='--help' or args[0]=='-h':
        print(usage)
        sys.exit()

    # Extract script parameters from options
    nbins       = int(options[0]['value'])
    sigma_min   = int(options[1]['value'])
    sigma_max   = int(options[2]['value'])
    includedreg = options[3]['value']
    excludedreg = options[4]['value']
    title       = options[5]['value']
    plotfile    = options[6]['value']

    # Plot significance distribution histogram
    plot_significance_distribution(args[0], nbins, sigma_min, sigma_max,
                                   includedreg, excludedreg, title,
                                   plotfile)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Plot and show significance distribution histogram
    show_significance_distribution()



#! /usr/bin/env python
# ==========================================================================
# Generates energy boundaries for stacked analysis
#
# Copyright (C) 2017-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import math
import gammalib
import ctools


# ============= #
# csebins class #
# ============= #
class csebins(ctools.csobservation):
    """
    Generates energy boundaries for stacked analysis
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)

        # Set members
        self._ebounds = gammalib.GEbounds()

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile
        """
        # Initialise observation container if it is currently empty
        if self.obs().size() == 0:

            # If an observation definition file was provided then load it,
            # otherwise build a single observation with response information
            if self['inobs'].is_valid():
                self.obs(gammalib.GObservations(self['inobs'].filename()))
            else:
                cta   = gammalib.GCTAObservation()
                caldb = gammalib.GCaldb('cta', self['caldb'].string())
                rsp   = gammalib.GCTAResponseIrf(self['irf'].string(), caldb)
                cta.response(rsp)
                self.obs().append(cta)

        # Query input parameters
        self['emin'].real()
        self['emax'].real()
        self['aeffthres'].real()
        self['bkgthres'].real()

        # Query ahead output model filename
        if self._read_ahead():
            self['outfile'].filename()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return

    def _insert_energy(self, responses, energy, comment):
        """
        Set energy boundaries for one CTA observation

        Parameters
        ----------
        responses : list of dict
            List of response dictionaries
        energy : float
            Energy to insert (TeV)
        comment : str
            Reason for energy insertion
        """
        # Set thresholds of all response components
        logE = math.log10(energy)
        for rsp in responses:
            rsp['thres'] = rsp['irf'](logE, 0.0, 0.0)

        # Convert energy into gammalib energy
        eng = gammalib.GEnergy(energy, 'TeV')
  
        # Set insertion message
        msg = str(energy)+' TeV'
        if len(comment) > 0:
            msg += ' ('+comment+')'
  
        # If energy boundaries are empty then append an energy interval
        # with zero size ...
        if self._ebounds.size() == 0:

            # Append energy interval with zero size
            self._ebounds.append(eng, eng)

            # Log first energy boundary
            self._log_value(gammalib.NORMAL, 'First boundary', msg)

        # ... otherwise append an interval
        elif self._ebounds.size() > 0:

            # If current boundaries have zero size then recover boundary energy,
            # remove boundary, and append a boundary of non-zero size
            if self._ebounds.emin() == self._ebounds.emax():

                # Recover old energy, remove boundary and append new boundary
                emin = self._ebounds.emin().copy() # Because emin becomes zero after remove
                emax = eng
                self._ebounds.remove(0)
                self._ebounds.append(emin, emax)

                # Log other energy boundary
                self._log_value(gammalib.NORMAL, 'Boundary', msg)

            else:

                # Use old emax as new emin. Only append if interval is non zero
                emin = self._ebounds.emax()
                if emin < eng:
                    self._ebounds.append(emin, eng)

                    # Log other energy boundary
                    self._log_value(gammalib.NORMAL, 'Boundary', msg)

        # Return
        return

    def _set_ebounds(self, obs):
        """
        Set energy boundaries for one CTA observation

        Parameters
        ----------
        obs : `~gammalib.GObservations`
            CTA observation
        """
        # Get parameters
        emin      = self['emin'].real()
        emax      = self['emax'].real()
        aeffthres = self['aeffthres'].real()
        bkgthres  = self['bkgthres'].real()

        # Build list of response dictionaries
        responses = []
  
        # Loop over all observations
        for run in obs:

            # Get response name
            mission    = run.response().caldb().mission()
            instrument = run.response().caldb().instrument()
            rspname    = run.response().rspname()
            name       = '%s::%s::%s' % (mission, instrument, rspname)

            # If response name exists already then examine next observation
            exists = False
            for rsp in responses:
                if rsp['name'] == name:
                    exists = True
                    break
            if exists:
                continue

            # Append effective area and background template to list
            aeff = {'name': name, 'type': 'Aeff',
                    'irf': run.response().aeff(), 'thres': 0.0}
            bkg  = {'name': name, 'type': 'Background',
                    'irf': run.response().background(), 'thres': 0.0}
            responses.append(aeff)
            responses.append(bkg)

            # Log response name
            self._log_value(gammalib.NORMAL, 'Append response', '%s (%s) [%s]' %
                            (rspname, instrument, mission))

        # Setup energy vector in log10(TeV)
        logEs = [math.log10(1.0e-3*float(i)) for i in range(int(emin*1000),int(emax*1000))]

        # Loop over all energies
        for logE in logEs:

            # Loop over all response components
            for rsp in responses:

                # Get IRF value
                irf = rsp['irf'](logE, 0.0, 0.0)

                # Get threshold
                if rsp['type'] == 'Aeff':
                    threshold = aeffthres
                else:
                    threshold = bkgthres

                # If threshold is zero and response is positive then insert
                # an energy ...
                if rsp['thres'] == 0.0 and irf > 0.0:
                    self._insert_energy(responses, math.pow(10.0, logE), rsp['type'])
                    break

                # ... otherwise if threshold is not zero then check whether
                # the fractional change exceeds the threshold
                elif rsp['thres'] != 0.0:
                    f = irf/rsp['thres']
                    if 1.0-f > threshold or f-1.0 > threshold:
                        self._insert_energy(responses, math.pow(10.0, logE), rsp['type'])
                        break

        # Insert energy at end of range
        self._insert_energy(responses, emax, 'End of range')

        # Log number of energy boundaries
        self._log_value(gammalib.NORMAL, 'Number of boundaries', self._ebounds.size())

        # Return
        return


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Write observation into logger
        self._log_observations(gammalib.NORMAL, self.obs(), 'Input observation')

        # Write header
        self._log_header1(gammalib.TERSE, 'Define energy boundaries')

        # Define energy boundaries for observation container
        self._set_ebounds(self.obs())

        # Return
        return

    def save(self):
        """
        Save energy boundaries
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Save energy boundaries')

        # Get energy boundaries filename
        outfile = self['outfile'].filename()

        # Log file name
        self._log_value(gammalib.NORMAL, 'Energy boundaries file', outfile.url())

        # Save energy boundaries
        self._ebounds.save(outfile, self._clobber())

        # Return
        return

    def ebounds(self):
        """
        Return energy boundaries

        Returns
        -------
        ebounds : `~gammalib.GEbounds()`
            Energy boundaries
        """
        # Return
        return self._ebounds


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csebins(sys.argv)

    # Execute application
    app.execute()



#!/usr/bin/env python
# ==========================================================================
# Dump information about observation into log file
#
# Copyright (C) 2015-2018 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools


# =============== #
# csobsinfo class #
# =============== #
class csobsinfo(ctools.csobservation):
    """
    Shows the content of an observation container
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor.
        """
        # Initialise application by calling the appropriate class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)

        # Initialise class members
        self._obj_dir        = None
        self._compute_offset = False
        self._offsets        = []
        self._zeniths        = []
        self._azimuths       = []
        self._pnt_ra         = []
        self._pnt_dec        = []
        self._ebounds        = gammalib.GEbounds()
        self._gti            = gammalib.GGti()

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile and setup the observation
        """
        # Get parameters
        if self.obs().size() == 0:
            self._require_inobs('csobsinfo::get_parameters')
            self.obs(self._get_observations(False))

        # Initialise object position
        self._obj_dir = gammalib.GSkyDir()

        # Get (optional) offset parameters
        self._compute_offset = self['offset'].boolean()
        if self._compute_offset:
            ra  = self['ra'].real()
            dec = self['dec'].real()
            self._obj_dir.radec_deg(ra,dec)

        # Read ahead DS9 filename
        if self._read_ahead():
            self['outds9file'].query()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Initialise arrays to store certain values for reuse
        # Todo, think about using a python dictionary
        self._offsets  = []
        self._zeniths  = []
        self._azimuths = []
        self._pnt_ra   = []
        self._pnt_dec  = []
        self._ebounds  = gammalib.GEbounds()
        self._gti      = gammalib.GGti()
        obs_names      = []

        # Initialise output to be filled
        ontime         = 0.0
        livetime       = 0.0
        n_events       = 0
        n_eventbins    = 0
        n_obs_binned   = 0
        n_obs_unbinned = 0

        # Write header
        if self._logTerse():
            self._log('\n')
            self._log.header1(gammalib.number('Observation', self.obs().size()))

        # Loop over observations
        for obs in self.obs():

            # Skip non-CTA observations
            if not obs.classname() == 'GCTAObservation':
                self._log('Skipping '+obs.instrument()+' observation\n')
                continue 

            # Use observed object as observation name if name is not given
            obs_name = obs.name()
            if obs_name == '':
                obs_name = obs.object()

            # Logging
            if self._logExplicit():
                obs_id   = obs.id()
                if obs_id != '':
                    log_name = obs_name + ' (ID='+obs_id+')'
                else:
                    log_name = obs_name
                self._log.header2(log_name)

            # Retrieve observation name
            obs_names.append(obs_name)

            # Retrieve energy boundaries
            obs_bounds = obs.events().ebounds()

            # Retrieve time interval
            obs_gti = obs.events().gti()

            # Compute mean time and dead time fraction in percent
            deadfrac = (1.0-obs.deadc())*100.0

            # Retrieve pointing and store Ra,Dec
            pnt_dir = obs.pointing().dir()
            self._pnt_ra.append(pnt_dir.ra_deg())
            self._pnt_dec.append(pnt_dir.dec_deg())

            # If avaliable append energy boundaries
            if obs_bounds.size() > 0 : 
                self._ebounds.append(obs_bounds.emin(),obs_bounds.emax())

            # Append time interval
            self._gti.append(obs_gti.tstart(), obs_gti.tstop())

            # Increment global livetime and ontime
            ontime   += obs.ontime()
            livetime += obs.livetime()

            # Bookkeeping
            if obs.eventtype() == 'CountsCube':
                n_eventbins  += obs.events().size()
                n_obs_binned += 1
                is_binned     = 'yes'
                is_what       = 'Number of bins'
            else:
                n_events       +=  obs.events().size()
                n_obs_unbinned += 1
                is_binned     = 'no'
                is_what       = 'Number of events'
            self._log_value(gammalib.EXPLICIT, 'Binned', is_binned)
            self._log_value(gammalib.EXPLICIT, is_what, obs.events().size())

            # Retrieve zenith and azimuth and store for later use
            zenith  = obs.pointing().zenith()
            azimuth = obs.pointing().azimuth()
            self._zeniths.append(zenith)
            self._azimuths.append(azimuth)

            # Optionally compute offset with respect to target direction
            if self._compute_offset:
                offset = pnt_dir.dist_deg(self._obj_dir)
                self._offsets.append(offset)

            # Optionally log details
            if self._logExplicit():

                # Log the observation energy range (if available)
                self._log.parformat('Energy range')
                if obs_bounds.size() == 0:
                    self._log('undefined')
                else:
                    self._log(str(obs_bounds.emin()))
                    self._log(' - ')
                    self._log(str(obs_bounds.emax()))
                self._log('\n')

                # Log observation time interval
                self._log.parformat('Time range (MJD)')
                if obs_gti.size() == 0:
                    self._log('undefined')
                else:
                    self._log(str(obs_gti.tstart().mjd()))
                    self._log(' - ')
                    self._log(str(obs_gti.tstop().mjd()))
                self._log('\n')

            # Log observation information
            self._log_value(gammalib.EXPLICIT, 'Ontime', '%.3f s' %
                            obs.ontime())
            self._log_value(gammalib.EXPLICIT, 'Livetime', '%.3f s' %
                            obs.livetime())
            self._log_value(gammalib.EXPLICIT, 'Deadtime fraction', '%.3f %%' %
                            deadfrac)
            self._log_value(gammalib.EXPLICIT, 'Pointing', pnt_dir)

            # Optionally log offset with respect to target direction
            if self._compute_offset:
                self._log_value(gammalib.EXPLICIT,
                                'Offset from target', '%.2f deg' % offset)

            # Log Zenith and Azimuth angles
            self._log_value(gammalib.EXPLICIT, 'Zenith angle', '%.2f deg' %
                            zenith)
            self._log_value(gammalib.EXPLICIT, 'Azimuth angle', '%.2f deg' %
                            azimuth)

        # Write summary header
        self._log_header1(gammalib.NORMAL, 'Summary')

        # Log general summary
        self._log_header3(gammalib.NORMAL, 'Observations')
        self._log_value(gammalib.NORMAL, 'Unbinned observations', n_obs_unbinned)
        self._log_value(gammalib.NORMAL, 'Binned observations', n_obs_binned)
        self._log_header3(gammalib.NORMAL, 'Events')
        self._log_value(gammalib.NORMAL, 'Number of events', n_events)
        self._log_value(gammalib.NORMAL, 'Number of bins', n_eventbins)

        # Compute mean offset, azimuth and zenith angle
        if len(self._offsets) > 0:
            mean_offset = '%.2f deg' % (sum(self._offsets) / len(self._offsets))
        else:
            mean_offset = 'Unknown'
        if len(self._zeniths) > 0:
            mean_zenith = '%.2f deg' % (sum(self._zeniths) / len(self._zeniths))
        else:
            mean_zenith = 'Unknown'
        if len(self._azimuths) > 0:
            mean_azimuth = '%.2f deg' % (sum(self._azimuths) / len(self._azimuths))
        else:
            mean_azimuth = 'Unknown'

        # Log mean offset, azimuth and zenith angle
        self._log_header3(gammalib.NORMAL, 'Pointings')
        self._log_value(gammalib.NORMAL, 'Mean offset angle',  mean_offset)
        self._log_value(gammalib.NORMAL, 'Mean zenith angle',  mean_zenith)
        self._log_value(gammalib.NORMAL, 'Mean azimuth angle', mean_azimuth)

        # Optionally log names of observations. Note that the set class is
        # used to extract all different observation names from the list of
        # observation names, and the set class is only available from
        # Python 2.4 on.
        if sys.version_info >= (2,4):
            obs_set = set(obs_names)
            for name in obs_set:
                self._log_value(gammalib.EXPLICIT,'"'+name+'"',
                                obs_names.count(name))

        # Get energy boundary information
        if self._ebounds.size() == 0:
            min_value = 'undefined'
            max_value = 'undefined'
        else:
            min_value = str(self._ebounds.emin())
            max_value = str(self._ebounds.emax())

        # Log energy range
        self._log_header3(gammalib.NORMAL, 'Energy range')
        self._log_value(gammalib.NORMAL, 'Minimum energy', min_value)
        self._log_value(gammalib.NORMAL, 'Maximum energy', max_value)

        # Log time range
        mjd = '%.3f - %.3f' % (self._gti.tstart().mjd(),  self._gti.tstop().mjd())
        utc = '%s - %s'     % (self._gti.tstart().utc(),  self._gti.tstop().utc())
        self._log_header3(gammalib.NORMAL, 'Time range')
        self._log_value(gammalib.NORMAL, 'MJD (days)', mjd)
        self._log_value(gammalib.NORMAL, 'UTC',  utc)

        # Log ontime and livetime in different units
        on_time   = '%.2f s = %.2f min = %.2f h' % \
                    (ontime, ontime/60., ontime/3600.)
        live_time = '%.2f s = %.2f min = %.2f h' % \
                    (livetime, livetime/60., livetime/3600.)
        self._log_value(gammalib.NORMAL, 'Total ontime', on_time)
        self._log_value(gammalib.NORMAL, 'Total livetime', live_time)

        # Return
        return

    def save(self):
        """ 
        Save pointings into DS9 region file

        This method saves all pointing directions that are found in the
        observation container into a DS9 region file. If "NONE" is
        specified for the "outds9file" parameter the method does nothing.
        """
        # Check if DS9 file is valid
        if self['outds9file'].is_valid():

            # Get output filename in case it was not read ahead
            ds9file = self['outds9file'].filename()

            # Write header
            self._log_header1(gammalib.TERSE, 'Save pointings in DS9 file')

            # Log filename
            self._log_value(gammalib.NORMAL, 'DS9 filename', ds9file.url())

            # Open file
            f = open(ds9file.url(),'w')

            # Write coordinate system
            f.write('fk5\n')

            # Loop over pointings
            for i in range(len(self._pnt_ra)):

                # Create string
                line  = 'point('
                line += str(self._pnt_ra[i])+','+str(self._pnt_dec[i])+')'
                line += ' # point=cross 20 width=3\n'

                # Write to file
                f.write(line)

            # Close file
            f.close()

        # Return
        return

    def zeniths(self):
        """
        Return zenith angles
        """
        return self._zeniths

    def azimuths(self):
        """
        Return azimuth angles
        """
        return self._azimuths

    def ras(self):
        """
        Return pointings right ascension
        """
        return self._pnt_ra

    def decs(self):
        """
        Return pointings declination
        """
        return self._pnt_dec

    def offsets(self):
        """
        Return offset angles
        """
        return self._offsets

    def ebounds(self):
        """
        Return energy boundaries
        """
        return self._ebounds

    def gti(self):
        """
        Return good time intervals
        """
        return self._gti


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csobsinfo(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# Display differential sensitivity plot generated by cssens
#
# Copyright (C) 2015-2017 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import math
import gammalib
import cscripts
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()


# ===================== #
# Read sensitivity data #
# ===================== #
def read_sensitivity(filename):
    """
    Read sensitivity information from CSV file

    Parameters
    ----------
    filename : str
        Name of CSV file

    Returns
    -------
    sensitivity : dict
        Dictionary with sensitivity information
    """
    # Read filename
    csv = gammalib.GCsv(filename,',')

    # Create dictionnary
    sensitivity = {}
    for column in range(csv.ncols()):
        name   = csv[0,column].rstrip('\r')
        values = []
        for row in range(csv.nrows()-1):
            values.append(float(csv[row+1,column]))
        sensitivity[name] = values

    # Check where file contains differential or integral sensitivity
    mode = 'Integral'
    if 'emax' in sensitivity:
        emax_ref = -1.0
        for value in sensitivity['emax']:
            if emax_ref < 0.0:
                emax_ref = value
            elif emax_ref != value:
                mode = 'Differential'
                break

    # Add mode
    sensitivity['mode'] = mode

    # Add linear energy values
    if mode == 'Differential':
        if 'loge' in sensitivity:
            name   = 'energy'
            values = []
            for value in sensitivity['loge']:
                values.append(math.pow(10.0, value))
            sensitivity[name] = values
    else:
        if 'emin' in sensitivity:
            name   = 'energy'
            values = []
            for value in sensitivity['emin']:
                values.append(value)
            sensitivity[name] = values

    # Return
    return sensitivity


# ===================== #
# Plot sensitivity data #
# ===================== #
def plot_sensitivity(sensitivity, filename, plotfile):
    """
    Plot sensitivity data

    Parameters
    ----------
    sensitivity : dict
        Dictionary with sensitivity information
    filename : str
        Name of CSV file
    plotfile : str
        Plot file name
    """
    # Build title
    title = sensitivity['mode'] + ' sensitivity ('+filename+')'

    # Create figure
    plt.figure()

    # Set plot attributes
    plt.title(title)
    plt.loglog()
    plt.grid()

    # Show differential sensitivity
    plt.plot(sensitivity['energy'], sensitivity['sensitivity'], 'ro-')

    # Set labels
    plt.xlabel('Energy (TeV)')
    if sensitivity['mode'] == 'Differential':
        plt.ylabel(r'E $\times$ F(E) (erg cm$^{-2}$ s$^{-1}$)')
    else:
        plt.ylabel(r'E $\times$ F($>$E) (erg cm$^{-2}$ s$^{-1}$)')

    # Show figure
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ================ #
# Show sensitivity #
# ================ #
def show_sensitivity():
    """
    Show sensitivity
    """
    # Set usage string
    usage = 'show_sensitivity.py [-p plotfile] [file]'

    # Set default options
    options = [{'option': '-p', 'value': ''}]

    # Get arguments and options from command line arguments
    args, options = cscripts.ioutils.get_args_options(options, usage)

    # Extract script parameters from options
    plotfile = options[0]['value']
    if len(args) > 0:
        filename = args[0]
    else:
        filename = 'sensitivity.dat'

    # Read sensitivity data
    sensitivity = read_sensitivity(filename)

    # Plot sensitivity data
    plot_sensitivity(sensitivity, filename, plotfile)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Show sensitivity
    show_sensitivity()



# ==========================================================================
# Python base class for unit tests
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import sys
import pickle
import gammalib


# ================================ #
# Python base class for unit tests #
# ================================ #
class test(gammalib.GPythonTestSuite):
    """
    Python base class for unit tests
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        gammalib.GPythonTestSuite.__init__(self)

        # Set test data directory
        self._datadir = os.environ['TEST_DATA']

        # Set some standard test data (additional test data will be set in
        # the derived classes as needed)
        self._events  = self._datadir + '/crab_events.fits'
        self._cntcube = self._datadir + '/crab_cntmap.fits'
        self._model   = self._datadir + '/crab.xml'
        self._caldb   = 'prod2'
        self._irf     = 'South_0.5h'

        # Return
        return

    # Set script name
    def _script(self, name):
        """
        Set script name

        Parameters
        ----------
        name : str
            Script name

        Returns
        -------
        script : str
            Script name including path
        """
        # Set script name dependent on whether the software is installed or
        # not. If the software is not installed the 'TEST_BUILDDIR' environment
        # variable will exist
        if 'TEST_BUILDDIR' in os.environ:
            script = '' + os.environ['TEST_BUILDDIR'] + \
                     '/cscripts/'+name+'.py'
        else:
            script = name

        # Return script name
        return script

    # Set tool name
    def _tool(self, name):
        """
        Set tool name

        Parameters
        ----------
        name : str
            Tool name

        Returns
        -------
        tool : str
            Tool name including path
        """
        # Set tool name dependent on whether the software is installed or
        # not. If the software is not installed the 'TEST_BUILDDIR' environment
        # variable will exist
        if 'TEST_BUILDDIR' in os.environ:
            tool = os.environ['TEST_BUILDDIR'] + '/src/'+name+'/'+name
        else:
            tool = name

        # Return tool name
        return tool

    # Execute command and catch any exception
    def _execute(self, cmd, success=True):
        """
        Execute command

        Parameters
        ----------
        cmd : str
            Command string to be executed
        success : bool
            True if success is expected, False otherwise

        Returns
        -------
        rc : int
            Return code
        """
        # Execute command, make sure we catch any exception
        try:
            rc = os.system(cmd+' > ctools_test_output.dmp 2>&1')
        except:
            rc = -1

        # If an exception occured then put output on screen
        if success and rc != 0 or not success and rc == 0:
            print('\n*** Unexpected exception (%d) for command %s:' % (rc, cmd))
            f = open('ctools_test_output.dmp', 'r')
            for line in f:
                print(line.rstrip('\n'))
            f.close()

        # Remove test output
        os.remove('ctools_test_output.dmp')

        # Return return code
        return rc

    # Return observation container with 3 identical event lists
    def _obs_events(self):
        """
        Return observation container with 3 identical event lists

        Returns
        -------
        obs : `~gammalib.GObservations`
            Observations container
        """
        # Load CTA event list
        cta = gammalib.GCTAObservation(self._events)

        # Allocate observation container
        obs = gammalib.GObservations()

        # Append CTA observation with ID "0001"
        cta.id('0001')
        obs.append(cta)

        # Append CTA observation with ID "0002"
        cta.id('0002')
        obs.append(cta)

        # Append CTA observation with ID "0003"
        cta.id('0003')
        obs.append(cta)

        # Return observation container
        return obs

    # Return observation container with event list, counts cube and LAT
    # observation
    def _obs_mixed(self):
        """
        Return observation container with event list, counts cube and LAT
        observation

        Returns
        -------
        obs : `~gammalib.GObservations`
            Observations container
        """
        # Load CTA event list
        obs1 = gammalib.GCTAObservation(self._events)
        obs1.id('0001')

        # Load CTA counts cube
        obs2 = gammalib.GCTAObservation(self._cntcube)
        obs2.id('0002')

        # Create empty LAT observation
        obs3 = gammalib.GLATObservation()
        obs3.id('0001')
        obs3.events(gammalib.GLATEventList())

        # Allocate observation container
        obs = gammalib.GObservations()

        # Append observations to container
        obs.append(obs1)
        obs.append(obs2)
        obs.append(obs3)

        # Return observation container
        return obs

    # Check
    def _check_help(self, tool):
        """
        Check for help text

        Parameters
        ----------
        name : str
            Tool name
        """
        # Setup tool --help option
        cmd = tool+' --help'

        # Check if execution was successful in case that the CTOOLS
        # environment variable was set or failed otherwise
        if 'CTOOLS' in os.environ:
            self.test_value(self._execute(cmd), 0,
                 'Check successful execution with --help option')
        else:
            self.test_assert(self._execute(cmd, success=False) != 0,
                 'Check execution failure with --help option')

        # Return
        return

    # Test pickeling
    def _pickeling(self, object):
        """
        Test class pickeling

        Parameters
        ----------
        object : `~gammalib.GBase`
            Class instance
        """
        # Get class name
        name = object.__class__.__name__
    
        # Start exception catching
        self.test_try('Test pickeling of "%s" class.' % name)

        # Test if pickling results in an exception
        try:
            dump       = pickle.dumps(object)
            obj        = pickle.loads(dump)
            dump_again = pickle.dumps(obj)
            self.test_assert(dump == dump_again,
                             'Check pickled dump of "%s\n%s\n%s"' %
                             (name, str(object), str(obj)))
            self.test_try_success()
        except:
            _, e, _ = sys.exc_info()
            self.test_try_failure('Error in pickeling "%s" (%s).' %
                                  (name, str(e)))

        # Return
        return obj



#! /usr/bin/env python
# ==========================================================================
# This scripts performs tests processing pipelines
#
# Copyright (C) 2014-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import ctools


# ============================ #
# Test class for ctobssim tool #
# ============================ #
class Test(gammalib.GPythonTestSuite):
    """
    Test class for pipelines
    """
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        gammalib.GPythonTestSuite.__init__(self)

        # Set test data directory
        self._datadir = os.environ['TEST_DATA']

        # Set some standard test data
        self._model   = self._datadir + '/crab.xml'
        self._caldb   = 'prod2'
        self._irf     = 'South_0.5h'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('pipelines')

        # Append tests
        self.append(self.test_unbinned_fits,
                    'Test unbinned pipeline with FITS file saving')
        self.append(self.test_unbinned_mem,
                    'Test unbinned in-memory pipeline')

        # Return
        return

    # Test unbinned pipeline with FITS file saving
    def test_unbinned_fits(self):
        """
        Test unbinned pipeline with FITS file saving
        """
        # Set script parameters
        events_name          = 'events.fits'
        selected_events_name = 'selected_events.fits'
        result_name          = 'results.xml'
        ra                   =   83.63
        dec                  =   22.01
        rad_sim              =    3.0
        rad_select           =    2.0
        tstart               =    0.0
        tstop                =  300.0
        emin                 =    1.0
        emax                 =  100.0

        # Simulate events
        sim = ctools.ctobssim()
        sim['inmodel']   = self._model
        sim['outevents'] = events_name
        sim['caldb']     = self._caldb
        sim['irf']       = self._irf
        sim['ra']        = ra
        sim['dec']       = dec
        sim['rad']       = rad_sim
        sim['tmin']      = tstart
        sim['tmax']      = tstop
        sim['emin']      = emin
        sim['emax']      = emax
        sim.execute()

        # Select events
        select = ctools.ctselect()
        select['inobs']  = events_name
        select['outobs'] = selected_events_name
        select['ra']     = ra
        select['dec']    = dec
        select['rad']    = rad_select
        select['tmin']   = tstart
        select['tmax']   = tstop
        select['emin']   = emin
        select['emax']   = emax
        select.execute()

        # Perform maximum likelihood fitting
        like = ctools.ctlike()
        like['inobs']    = selected_events_name
        like['inmodel']  = self._model
        like['outmodel'] = result_name
        like['caldb']    = self._caldb
        like['irf']      = self._irf
        like.execute()

        # Return
        return

    # Test unbinned in-memory pipeline
    def test_unbinned_mem(self):
        """
        Test unbinned in-memory pipeline
        """
        # Set script parameters
        ra         =   83.63
        dec        =   22.01
        rad_sim    =    3.0
        rad_select =    2.0
        tstart     =    0.0
        tstop      =  300.0
        emin       =    1.0
        emax       =  100.0

        # Simulate events
        sim = ctools.ctobssim()
        sim['inmodel'] = self._model
        sim['caldb']   = self._caldb
        sim['irf']     = self._irf
        sim['ra']      = ra
        sim['dec']     = dec
        sim['rad']     = rad_sim
        sim['tmin']    = tstart
        sim['tmax']    = tstop
        sim['emin']    = emin
        sim['emax']    = emax
        sim.run()

        # Select events
        select = ctools.ctselect(sim.obs())
        select['ra']   = ra
        select['dec']  = dec
        select['rad']  = rad_select
        select['tmin'] = tstart
        select['tmax'] = tstop
        select['emin'] = emin
        select['emax'] = emax
        select.run()

        # Perform maximum likelihood fitting
        like = ctools.ctlike(select.obs())
        like.run()

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csmodelmerge script.
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# ================================== #
# Test class for csmodelmerge script #
# ================================== #
class Test(test):
    """
    Test class for csmodelmerge script

    This test class makes unit tests for the csmodelmerge script by using it
    from the command line and from Python.
    """
    
    # Constructor
    def __init__(self):
        """
        Constructor.
        """
        # Call base class constructor
        test.__init__(self)

        # Set members
        self._model1  = self._datadir + '/crab.xml'
        self._model2  = self._datadir + '/model_cube_background1.xml'
        self._model3  = self._datadir + '/model_cube_background*.xml'
        self._model4  = '@' + self._datadir + '/models.txt'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions.
        """
        # Set test name
        self.name('csmodelmerge')

        # Append tests
        self.append(self._test_cmd, 'Test csmodelmerge on command line')
        self.append(self._test_python, 'Test csmodelmerge from Python')

        # Return
        return

    # Test csmodelmerge on command line
    def _test_cmd(self):
        """
        Test csmodelmerge on the command line
        """
        # Set script name
        csmodelmerge = self._script('csmodelmerge')

        # Setup csmodelmerge command
        cmd = csmodelmerge+' inmodels="'+self._model1+' '+self._model2+'"'+ \
                           ' outmodel="csmodelmerge_cmd1.xml"'+ \
                           ' logfile="csmodelmerge_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check model file
        self._check_model_file('csmodelmerge_cmd1.xml', 3)

        # Setup csmodelmerge command
        cmd = csmodelmerge+' inmodels="model_that_does_not_exist.xml"'+ \
                           ' outmodel="csmodelmerge_cmd2.xml"'+ \
                           ' logfile="csmodelmerge_cmd2.log" debug=yes'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csmodelmerge --help
        self._check_help(csmodelmerge)

        # Return
        return

    # Test csmodelmerge from Python
    def _test_python(self):
        """
        Test csmodelmerge from Python
        """
        # Set-up csmodelmerge for space-separated input
        modelmerge = cscripts.csmodelmerge()
        modelmerge['inmodels'] = self._model1+' '+self._model2
        modelmerge['outmodel'] = 'csmodelmerge_py1.xml'
        modelmerge['logfile']  = 'csmodelmerge_py1.log'
        modelmerge['chatter']  = 2

        # Run csmodelmerge script and save models
        modelmerge.logFileOpen()   # Make sure we get a log file
        modelmerge.run()
        modelmerge.save()

        # Check model file
        self._check_model_file('csmodelmerge_py1.xml', 3)

        # Set-up csmodelmerge for semi-colon separated input
        modelmerge = cscripts.csmodelmerge()
        modelmerge['inmodels'] = self._model1+';'+self._model2
        modelmerge['outmodel'] = 'csmodelmerge_py2.xml'
        modelmerge['logfile']  = 'csmodelmerge_py2.log'
        modelmerge['chatter']  = 3

        # Execute csmodelmerge script
        modelmerge.execute()

        # Check model file
        self._check_model_file('csmodelmerge_py2.xml', 3)

        # Set-up csmodelmerge for wildcard input
        modelmerge = cscripts.csmodelmerge()
        modelmerge['inmodels'] = self._model3
        modelmerge['outmodel'] = 'csmodelmerge_py3.xml'
        modelmerge['logfile']  = 'csmodelmerge_py3.log'
        modelmerge['chatter']  = 4

        # Execute csmodelmerge script
        modelmerge.execute()

        # Check model file
        self._check_model_file('csmodelmerge_py3.xml', 2)

        # Set-up csmodelmerge for ASCII file
        modelmerge = cscripts.csmodelmerge()
        modelmerge['inmodels'] = self._model4
        modelmerge['outmodel'] = 'csmodelmerge_py4.xml'
        modelmerge['logfile']  = 'csmodelmerge_py4.log'
        modelmerge['chatter']  = 4

        # Execute csmodelmerge script
        modelmerge.execute()

        # Check model file
        self._check_model_file('csmodelmerge_py4.xml', 2)

        # Return
        return

    # Check model file
    def _check_model_file(self, filename, number):
        """
        Check model file
        """
        # Open models
        models = gammalib.GModels(filename)

        # Check number of models
        self.test_value(models.size(), number,
                        'Check for '+str(number)+' models in XML file')
        
        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the ctulimit tool.
#
# Copyright (C) 2015-2018 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import ctools
from testing import test


# ============================ #
# Test class for ctulimit tool #
# ============================ #
class Test(test):
    """
    Test class for ctulimit tool
    """
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('ctulimit')

        # Append tests
        self.append(self._test_cmd, 'Test ctulimit on command line')
        self.append(self._test_python, 'Test ctulimit from Python')

        # Return
        return

    # Test ctulimit on command line
    def _test_cmd(self):
        """
        Test ctulimit on the command lines
        """
        # Set tool name
        ctulimit = self._tool('ctulimit')

        # Setup ctulimit command
        cmd = ctulimit+' inobs="'+self._events+'"'+ \
                       ' inmodel="'+self._model+'" srcname="Crab"'+ \
                       ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                       ' tol=0.1 logfile="ctulimit_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Setup ctulimit command
        cmd = ctulimit+' inobs="event_file_that_does_not_exist.fits"'+ \
                       ' inmodel="'+self._model+'" srcname="Crab"'+ \
                       ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                       ' tol=0.1 logfile="ctulimit_cmd2.log" debug=yes chatter=2'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Setup ctulimit command
        cmd = ctulimit+' inobs="'+self._events+'"'+ \
                       ' inmodel="'+self._model+'" srcname="Vela"'+ \
                       ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                       ' tol=0.1 logfile="ctulimit_cmd2.log" debug=yes chatter=2'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid source name when executed from command line')

        # Check ctulimit --help
        self._check_help(ctulimit)

        # Return
        return

    # Test ctulimit from Python
    def _test_python(self):
        """
        Test ctulimit from Python
        """
        # Allocate ctulimit
        ulimit = ctools.ctulimit()

        # Check that empty ctulimit tool holds zero upper limits
        self.test_value(ulimit.diff_ulimit(), 0.0, 1.0e-21,
                        'Check differential upper limit')
        self.test_value(ulimit.flux_ulimit(), 0.0, 1.0e-16,
                        'Check upper limit on photon flux')
        self.test_value(ulimit.eflux_ulimit(), 0.0, 1.0e-16,
                        'Check upper limit on energy flux')

        # Check that saving does not nothing
        ulimit['logfile'] = 'ctulimit_py0.log'
        ulimit.logFileOpen()
        ulimit.save()

        # Check that clearing does not lead to an exception or segfault
        ulimit.clear()

        # Now set ctulimit parameters
        ulimit['inobs']   = self._events
        ulimit['inmodel'] = self._model
        ulimit['srcname'] = 'Crab'
        ulimit['caldb']   = self._caldb
        ulimit['irf']     = self._irf
        ulimit['tol']     = 0.1
        ulimit['logfile'] = 'ctulimit_py1.log'
        ulimit['chatter'] = 2

        # Run ctulimit tool
        ulimit.logFileOpen()   # Make sure we get a log file
        ulimit.run()
        ulimit.save()

        # Set reference value
        ref_diff  = 2.75359655874408e-17
        ref_flux  = 1.66942609234064e-11
        ref_eflux = 6.4588860064421e-11

        # Check results
        self.test_value(ulimit.diff_ulimit(), ref_diff, 1.0e-21,
                        'Check differential upper limit')
        self.test_value(ulimit.flux_ulimit(), ref_flux, 1.0e-16,
                        'Check upper limit on photon flux')
        self.test_value(ulimit.eflux_ulimit(), ref_eflux, 1.0e-16,
                        'Check upper limit on energy flux')

        # Check obs() method
        self.test_value(ulimit.obs().size(), 1,
                        'Check number of observations in container')

        # Check opt() method
        self.test_value(ulimit.opt().status(), 0, 'Check optimizer status')

        # Copy ctulimit tool
        cpy_ulimit = ulimit.copy()

        # Check results of copy
        self.test_value(cpy_ulimit.diff_ulimit(), ref_diff, 1.0e-21,
                        'Check differential upper limit')
        self.test_value(cpy_ulimit.flux_ulimit(), ref_flux, 1.0e-16,
                        'Check upper limit on photon flux')
        self.test_value(cpy_ulimit.eflux_ulimit(), ref_eflux, 1.0e-16,
                        'Check upper limit on energy flux')

        # Check results
        self.test_value(cpy_ulimit.diff_ulimit(), ref_diff, 1.0e-21,
                        'Check differential upper limit')
        self.test_value(cpy_ulimit.flux_ulimit(), ref_flux, 1.0e-16,
                        'Check upper limit on photon flux')
        self.test_value(cpy_ulimit.eflux_ulimit(), ref_eflux, 1.0e-16,
                        'Check upper limit on energy flux')

        # Now clear copy of ctulimit tool
        cpy_ulimit.clear()

        # Check that empty ctulimit tool holds zero upper limits
        self.test_value(cpy_ulimit.diff_ulimit(), 0.0, 1.0e-21,
                        'Check differential upper limit')
        self.test_value(cpy_ulimit.flux_ulimit(), 0.0, 1.0e-16,
                        'Check upper limit on photon flux')
        self.test_value(cpy_ulimit.eflux_ulimit(), 0.0, 1.0e-16,
                        'Check upper limit on energy flux')

        # Run ctlike to get an initial log-likelihood solution
        like = ctools.ctlike()
        like['inobs']    = self._events
        like['inmodel']  = self._model
        like['caldb']    = self._caldb
        like['irf']      = self._irf
        like.run()

        # Now set ctulimit tool using the observation container from the
        # previous run. This should avoid the necessity to recompute the
        # maximum likelihood
        ulimit = ctools.ctulimit(like.obs())
        ulimit['srcname'] = 'Crab'
        ulimit['tol']     = 0.1
        ulimit['logfile'] = 'ctulimit_py2.log'
        ulimit['chatter'] = 3

        # Execute ctulimit tool
        ulimit.logFileOpen()  # Needed to get a new log file
        ulimit.execute()

        # Check results
        self.test_value(ulimit.diff_ulimit(), ref_diff, 1.0e-21,
                        'Check differential upper limit')
        self.test_value(ulimit.flux_ulimit(), ref_flux, 1.0e-16,
                        'Check upper limit on photon flux')
        self.test_value(ulimit.eflux_ulimit(), ref_eflux, 1.0e-16,
                        'Check upper limit on energy flux')

        # Test invalid model name
        ulimit['srcname'] = 'Weihnachtsstern'
        ulimit['logfile'] = 'ctulimit_py3.log'
        ulimit.logFileOpen()
        self.test_try('Test invalid model name')
        try:
            ulimit.execute()
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test specification of background model
        ulimit['srcname'] = 'Background'
        ulimit['logfile'] = 'ctulimit_py4.log'
        ulimit.logFileOpen()
        self.test_try('Test invalid model name')
        try:
            ulimit.execute()
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test run with too few iterations
        ulimit['srcname']  = 'Crab'
        ulimit['max_iter'] = 1
        ulimit['logfile']  = 'ctulimit_py5.log'
        ulimit.logFileOpen()
        self.test_try('Test ctulimit with too few iterations')
        try:
            ulimit.execute()
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the ctprob tool.
#
# Copyright (C) 2017-2018 by Leonardo Di Venere
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import ctools
from testing import test


# ============================ #
# Test class for ctprob tool #
# ============================ #
class Test(test):
    """
    Test class for ctprob tool
    """
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set test data
        self._invalid_events = self._datadir + '/invalid_event_list.fits'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('ctprob')

        # Append tests
        self.append(self._test_cmd, 'Test ctprob on command line')
        self.append(self._test_python, 'Test ctprob from Python')

        # Return
        return

    # Test ctprob on command line
    def _test_cmd(self):
        """
        Test ctprob on the command line
        """
        # Set tool name
        ctprob = self._tool('ctprob')

        # Setup ctprob command
        cmd = ctprob+' inobs="'+self._events+'"'+ \
                     ' outobs="ctprob_cmd1.fits"'+ \
                     ' inmodel='+self._model+ \
                     ' caldb='+self._caldb+' irf='+self._irf+ \
                     ' edisp=no'+ \
                     ' logfile="ctprob_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check result file
        self._check_result_file('ctprob_cmd1.fits')

        # Setup ctprob command
        cmd = ctprob+' inobs="event_file_that_does_not_exist.fits"'+ \
                     ' outobs="ctprob_cmd2.fits"'+ \
                     ' inmodel='+self._model+ \
                     ' caldb='+self._caldb+' irf='+self._irf+ \
                     ' edisp=no'+ \
                     ' logfile="ctprob_cmd2.log" debug=yes chatter=1'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check ctprob --help
        self._check_help(ctprob)

        # Return
        return

    # Test ctprob from Python
    def _test_python(self):
        """
        Test ctprob from Python
        """
        # Allocate empty ctprob tool
        prob = ctools.ctprob()

        # Check that empty ctprob tool holds an empty observation
        self._check_obs(prob.obs(), nobs=0)

        # Check that saving does nothing
        prob['outobs']  = 'ctprob_py0.fits'
        prob['logfile'] = 'ctprob_py0.log'
        prob.logFileOpen()
        prob.save()
        self.test_assert(not os.path.isfile('ctprob_py0.fits'),
             'Check that no event list has been created')

        # Check that clearing does not lead to an exception or segfault
        prob.clear()

        # Now set ctprob parameters
        prob['inobs']   = self._events
        prob['outobs']  = 'ctprob_py1.fits'
        prob['inmodel'] = self._model
        prob['caldb']   = self._caldb
        prob['irf']     = self._irf
        prob['edisp']   = False
        prob['publish'] = True
        prob['logfile'] = 'ctprob_py1.log'
        prob['chatter'] = 2

        # Run ctprob tool
        prob.logFileOpen()   # Make sure we get a log file
        prob.run()
        prob.save()

        # Check result file
        self._check_result_file('ctprob_py1.fits')

        # Setup ctprob tool for an invalid event file
        prob = ctools.ctprob()
        prob['inobs']   = self._invalid_events
        prob['outobs']  = 'ctprob_py2.fits'
        prob['inmodel'] = self._model
        prob['caldb']   = self._caldb
        prob['irf']     = self._irf
        prob['edisp']   = False
        prob['logfile'] = 'ctprob_py2.log'
        prob['chatter'] = 3

        # Execute tool
        prob.logFileOpen()  # Needed to get a new log file
        self.test_try('Test invalid event file')
        try:
            prob.execute()
            self.test_try_failure('Exception not thrown')
        except RuntimeError:
            self.test_try_success()

        # Setup ctprob with a list of observations and test prefix
        prob = ctools.ctprob()
        prob['inobs']   = self._datadir + '/obs_unbinned.xml'
        prob['prefix']  = 'ctprob_py3_'
        prob['outobs']  = 'ctprob_py3_obs.xml'
        prob['inmodel'] = self._model
        prob['caldb']   = self._caldb
        prob['irf']     = self._irf
        prob['edisp']   = True
        prob['logfile'] = 'ctprob_py3.log'
        prob['chatter'] = 4

        # Run ctprob tool
        prob.logFileOpen()   # Make sure we get a log file
        prob.execute()

        # Check result file
        self._check_result_file('ctprob_py3_crab_events.fits')

        # Copy ctprob tool
        cpy_prob = prob.copy()

        # Clear phase tool
        prob.clear()

        # Execute copy
        cpy_prob['prefix']  = 'ctprob_py4_'
        cpy_prob['outobs']  = 'ctprob_py4_obs.xml'
        cpy_prob['logfile'] = 'ctprob_py4.log'
        cpy_prob['chatter'] = 4
        cpy_prob.logFileOpen()  # Needed to get a new log file
        cpy_prob.execute()

        # Check result file
        self._check_result_file('ctprob_py4_ctprob_py3_crab_events.fits')

        # Return
        return

    # Check result file
    def _check_result_file(self, filename, nevents=253):
        """
        Check result file

        Parameters
        ----------
        filename : str
            Event list file name
        nevents : int, optional
            Expected number of events
        """
        # Open result file
        events = gammalib.GCTAEventList(filename)

        # Check event list
        self._check_events(events, nevents=nevents)

        # Check that probability columns exist
        colname1 = 'PROB_Background'
        colname2 = 'PROB_Crab'
        self._check_column(filename, colname1)
        self._check_column(filename, colname2)        

        # Check that the probability values are correctly normalized
        self._check_normalization(filename, [colname1,colname2])

        # Return
        return

    # Check observation and event list
    def _check_obs(self, obs, nobs=1, nevents=253):
        """
        Check observation and event list

        Parameters
        ----------
        obs : `~gammalib.GObservations`
            Models
        nobs : int, optional
            Expected number of observations
        nevents : int, optional
            Expected number of events
        """
        # Check number of observations
        self.test_value(obs.size(), nobs, 'Check number of observations')

        # If there is an observation then check the event list of the first
        # one
        if obs.size() > 0:
            self._check_events(obs[0].events(), nevents=nevents)

        # Return
        return

    # Check events
    def _check_events(self, events, nevents=253):
        """
        Check event list

        Parameters
        ----------
        events : `~gammalib.GCTAEventList`
            Event list
        nevents : int, optional
            Expected number of events
        """
        # Check event list
        self.test_value(events.size(), nevents, 'Check number of events')

        # Return
        return


    def _check_column(self, filename, colname):
        """
        Check that column colname exists.

        Parameters
        ----------
        filename : str
            Event list file name
        colname : str
            Name of the column to check
        """
        # Check that column exists
        gfits = gammalib.GFits(filename)
        gtable = gfits[1]
        check = int(gtable.contains(colname))
        self.test_value(check, 1, 'Check column '+colname)

        # Return
        return


    def _check_normalization(self, filename, colnames):
        """
        Check that the probability values in columns identified by colnames are
        correctly normalized.

        Parameters
        ----------
        filename : str
            Event list file name
        colnames : list
            Names of the columns to check
        """
        # Check that the probability values are correctly normalized

        # Open fits table
        gfits  = gammalib.GFits(filename)
        gtable = gfits[1]
        ncols  = gtable.ncols()
        nevt   = gtable.nrows()
        srcs   = []

        # Loop over colnames
        for colname in colnames:

            # Get the correct column corresponding to colname from the table
            for icol in range(ncols):
                gcol = gtable[icol] 
                if gcol.name()==colname:
                    break

            # Read and save column content 
            src1 = []
            for ievt in range(nevt):
                src1.append(gcol[ievt])
            srcs.extend(src1)

        # Compute sum
        tot = sum(srcs)
        self.test_value(int(tot+0.5), nevt, 'Check that probability columns '
                                            'are normalized')

        # Return
        return




#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csmodelsois tool.
#
# Copyright (C) 2017-2018 Josh Cardenzana
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import cscripts
from testing import test


# =============================== #
# Test class for csmodelsois tool #
# =============================== #
class Test(test):
    """
    Test class for csmodelsois tool

    This test class makes unit tests for the csmodelsois tool by using it from
    the command line and from Python.
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('csmodelsois')

        # Append tests
        self.append(self._test_cmd, 'Test csmodelsois on command line')
        self.append(self._test_python, 'Test csmodelsois from Python')

        # Return
        return

    # Test csmodelsois on command line
    def _test_cmd(self):
        """
        Test csmodelsois on the command line
        """
        # Set tool name
        csmodelsois = self._script('csmodelsois')

        # Setup csmodelsois command
        cmd = csmodelsois+' inmodel="'+self._model+'"'+ \
                          ' outcube="csmodelsois_cmd1.fits"'+\
                          ' emin=0.1 emax=100.0 enumbins=10 ebinalg=LOG'+ \
                          ' nxpix=100 nypix=100 binsz=0.04 coordsys=CEL'+ \
                          ' soilist="" outmodel=NONE'+ \
                          ' ra=83.63 dec=22.01 proj=CAR'+ \
                          ' logfile="csmodelsois_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check map cube
        self._check_result_file('csmodelsois_cmd1.fits')

        # Setup csmodelsois command
        cmd = csmodelsois+' inmodel="events_that_do_not_exist.fits"'+ \
                          ' outcube="ccsmodelsois_cmd2.fits"'+\
                          ' emin=0.1 emax=100.0 enumbins=10 ebinalg=LOG'+ \
                          ' nxpix=100 nypix=100 binsz=0.04 coordsys=CEL'+ \
                          ' soilist="" outmodel=NONE'+ \
                          ' ra=83.63 dec=22.01 proj=CAR'+ \
                          ' logfile="csmodelsois_cmd2.log" debug=yes'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csmodelsois --help
        self._check_help(csmodelsois)

        # Return
        return

    # Test csmodelsois from Python
    def _test_python(self):
        """
        Test csmodelsois from Python
        """
        # Allocate csmodelsois
        modelsois = cscripts.csmodelsois()

        # Check that empty csmodelsois tool holds an map cube that has no
        # energy bins
        self._check_cube(modelsois.mapcube(), nmaps=0, npixels=0)

        # Check that saving does nothing
        modelsois['outcube']  = 'csmodelsois_py0.fits'
        modelsois['outmodel'] = 'NONE'
        modelsois['logfile']  = 'csmodelsois_py0.log'
        modelsois.logFileOpen()
        modelsois.save()
        self.test_assert(not os.path.isfile('csmodelsois_py0.fits'),
             'Check that no map cube has been created')

        # Check that clearing does not lead to an exception or segfault
        modelsois.clear()

        # Now set csmodelsois parameters
        modelsois['inmodel']  = self._model
        modelsois['ebinalg']  = 'LOG'
        modelsois['emin']     = 0.1
        modelsois['emax']     = 100.0
        modelsois['enumbins'] = 10
        modelsois['nxpix']    = 100
        modelsois['nypix']    = 100
        modelsois['binsz']    = 0.04
        modelsois['coordsys'] = 'CEL'
        modelsois['proj']     = 'CAR'
        modelsois['ra']       = 83.63
        modelsois['dec']      = 22.01
        modelsois['ptsrcsig'] = 1.0
        modelsois['outcube']  = 'csmodelsois_py1.fits'
        modelsois['logfile']  = 'csmodelsois_py1.log'
        modelsois['chatter']  = 2
        modelsois['soilist']  = ''
        modelsois['outmodel'] = 'NONE'

        # Run csmodelsois tool
        modelsois.logFileOpen()   # Make sure we get a log file
        modelsois.run()

        # Save map cube
        modelsois.save()

        # Check map cube
        self._check_result_file('csmodelsois_py1.fits')

        # Allocate csmodelsois scripts, set models and do now a linear
        # binning; also do not use Gaussian point sources
        modelsois = cscripts.csmodelsois()
        modelsois['inmodel']  = self._model
        modelsois['ebinalg']  = 'LIN'
        modelsois['emin']     = 0.1
        modelsois['emax']     = 100.0
        modelsois['enumbins'] = 10
        modelsois['nxpix']    = 100
        modelsois['nypix']    = 100
        modelsois['binsz']    = 0.04
        modelsois['coordsys'] = 'CEL'
        modelsois['proj']     = 'CAR'
        modelsois['ra']       = 83.63
        modelsois['dec']      = 22.01
        modelsois['ptsrcsig'] = 0.0
        modelsois['outcube']  = 'csmodelsois_py2.fits'
        modelsois['logfile']  = 'csmodelsois_py2.log'
        modelsois['chatter']  = 4
        modelsois['soilist']  = ''
        modelsois['outmodel'] = 'NONE' 

        # Execute mapcube
        modelsois.logFileOpen()  # Needed to get a new log file
        modelsois.execute()

        # Check result file
        self._check_result_file('csmodelsois_py2.fits')

        # Update output filenames and output a new model file
        modelsois['outcube']  = 'csmodelsois_py3.fits'
        modelsois['logfile']  = 'csmodelsois_py3.log'
        modelsois['outmodel'] = 'csmodelsois_py3.xml'

        # Execute the file
        modelsois.logFileOpen()
        modelsois.execute()

        # Now check that the output model file doesnt contain a Crab Sources
        models3 = gammalib.GModels('csmodelsois_py3.xml')
        self.test_assert(not models3.contains('Crab'),
                         'Check Crab model is not present')
        self.test_assert(models3.contains(modelsois.cubemodelname()), 
                         'Check cube model is present')
        self._check_result_file('csmodelsois_py3.fits')

        # Return
        return

    # Check result file
    def _check_result_file(self, filename):
        """
        Check content of map cube

        Parameters
        ----------
        filename : str
            Map cube file name
        """
        # Load map cube
        cube = gammalib.GModelSpatialDiffuseCube(filename)

        # Check map cube
        self._check_cube(cube)

        # Return
        return

    # Check map cube
    def _check_cube(self, cube, nmaps=11, npixels=10000):
        """
        Check map cube

        Parameters
        ----------
        cube : `~gammalib.GModelSpatialDiffuseCube`
            Map cube
        nmaps : int, optional
            Number of maps
        npixels : int, optional
            Number of pixels
        """
        # Get energies (this forces loading in case the map cube is not
        # loaded)
        self.test_value(len(cube.energies()), nmaps, 'Check number of energies')
        
        # Check dimensions
        self.test_value(cube.maps(), nmaps, 'Check number of maps')
        self.test_value(cube.pixels(), npixels, 'Check number of pixels')

        # Return
        return



#!/usr/bin/env python
# ==========================================================================
# Inspect an IACT data storage
#
# Copyright (C) 2015-2018 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import os
import json
import gammalib
import ctools


# ================ #
# csiactdata class #
# ================ #
class csiactdata(ctools.cscript):
    """
    Inspect an IACT data storage
    
    This script inspects the available FITS data storage on the user machine
    and writes information about the available FITS configurations into a log
    file or prints it on the screen. These information can be used as input
    for 'csiactobs'.
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor.
        """
        # Initialise application by calling the base class constructor
        self._init_cscript(self.__class__.__name__, ctools.__version__, argv)

        # Set name
        self._datapath    = os.getenv('VHEFITS','')
        self._prodnames   = []
        self._master_indx = ''
        self._master_file = ''

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile and setup the observation
        """
        # Query datapath. If the parameter is not NONE then use it, otherwise
        # use the datapath from the VHEFITS environment variable
        datapath = self['datapath'].string()
        if gammalib.toupper(datapath) != 'NONE':
            self._datapath = datapath
        else:
            self._datapath = os.getenv('VHEFITS','')

        # Expand environment
        self._datapath = gammalib.expand_env(self._datapath)

        # Get filename of master index file
        self._master_indx = self['master_indx'].string()
        self._master_file = os.path.join(self._datapath, self._master_indx)

        # Check for presence of master index file
        if not os.path.isfile(self._master_file):
            raise RuntimeError('Master index file "'+self._master_file+
                               '" not found. Use hidden parameter '+
                               '"master_indx" to specifiy a different '+
                               'filename.')

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)
        
        # Return
        return

    # Log configuration
    def _log_configuration(self, config):
        """
        Log configuration
        
        Parameters
        ----------
        config : dict
            Dictionary containing data configuration
        """
        # Log configuration information
        self._log_header3(gammalib.TERSE, str(config['name']))
        self._log_value(gammalib.TERSE, 'Name', config['name'])
        self._log_value(gammalib.TERSE, 'Observation index', config['obsindx'])
        self._log_value(gammalib.TERSE, 'HDU index', config['hduindx'])
            
        # Print additional information
        for key in config:
                
            # Skip keys that were treated above
            if key in ['name','hduindx','obsindx']:
                continue
            self._log_value(gammalib.TERSE, key, config[key])

        # Return
        return


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Write header into logger
        if self._logTerse():
            self._log('\n')
            self._log.header1('Data storage entries')
            self._log.parformat('Master index file')
            self._log(self._master_file)
            self._log('\n')
        
        # Open and load JSON file
        json_data = open(self._master_file).read()
        data      = json.loads(json_data)    
        configs   = data['datasets']
        
        # Initialise array for available names
        self._prodnames = []

        # Write header into logger
        self._log('\n')
        self._log.header2('Available data configs')
        
        # Loop over configs and log available configs       
        for config in configs:       
            
            # Create hdu and obs index files
            hdu          = os.path.join(self._datapath, config['hduindx'])
            obs          = os.path.join(self._datapath, config['obsindx'])
            filename_hdu = gammalib.GFilename(str(hdu)+'[HDU_INDEX]')
            filename_obs = gammalib.GFilename(str(obs)+'[OBS_INDEX]')
            
            # If index files are available then log configuration
            # information ...
            if (filename_hdu.is_fits() and filename_obs.is_fits()):

                # Append to available names
                self._prodnames.append(str(config['name']))
                
                # Log information
                self._log_configuration(config)

            # ... otherwise log that the configuration is not available
            else:
                if self._logTerse():
                    self._log.header3(str(config['name']))
                    self._log(' Not available\n')

        # Return
        return       

    def names(self):
        """ 
        Return available FITS production names
        
        Returns
        -------
        names : list of str
            List of available FITS production names.
        """
        # Return
        return self._prodnames


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csiactdata(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# Generation of an observation definition file
#
# Copyright (C) 2015-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools


# ============== #
# csobsdef class #
# ============== #
class csobsdef(ctools.cscript):
    """
    Creates an observation definition file from a pointing list
    
    The pointing list is a comma-separated value ASCII file with header
    keywords in the first row followed by a list of pointings (one
    pointing per row). The following header keywords are supported (case
    sensitive, column order irrelevant):

    name     - Observation name string
    id       - Unique observation identifier string
    ra       - Right Ascension of pointing (deg)
    dec      - Declination of pointing (deg)
    lon      - Galactic longitude of pointing (deg)
    lat      - Galactic latitude of pointing (deg)
    tmin     - Start of pointing (seconds)
    duration - Duration of pointing (seconds)
    emin     - Lower energy limit (TeV)
    emax     - Upper energy limit (TeV)
    rad      - Radius of region of interest (deg)
    deadc    - Deadtime correction factor [0-1]
    caldb    - Calibration database
    irf      - Response function name

    Only the pairs (ra,dec) or (lon,lat) are mandatory header keywords.
    All other keywords are optional and can be specified when calling
    csobsdef as user parameters. The only exception is the "duration"
    keyword that will automatically be queried.

    Examples:

        ./csobsdef
            Creates minimal observation definition file.

        ./csobsdef emin=0.1 emax=100.0
            Creates observation definition file with an energy range
            100 GeV - 100 TeV.

        ./csobsdef rad=5
            Creates observation definition file with a ROI radius of
            5 degrees.

        ./csobsdef caldb=prod2 irf=South_50h
            Creates observation definition file using the "South_50h"
            IRF in the "prod2" calibration database.
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor

        Parameters
        ----------
        argv : list of str
            List of IRAF command line parameter strings of the form
            ``parameter=3``.
        """
        # Initialise application by calling the base class constructor
        self._init_cscript(self.__class__.__name__, ctools.__version__, argv)

        # Initialise class members
        self._obs    = gammalib.GObservations()
        self._pntdef = gammalib.GCsv()
        self._tmin   = 0.0

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile
        """
        # Query input filename if necessary
        if self._pntdef.size() == 0:
            self['inpnt'].filename()

        # Read ahead parameters
        if self._read_ahead():
            self['outobs'].query()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return

    def _set_irf(self, obs, caldb, irf):
        """
        Set response for an observation
        
        The method creates an XML element so that that the response XML
        writer will write the database and response name into the
        observation definition file.

        Parameters
        ----------
        obs : `~gammalib.GCTAObservation`
            CTA observation
        caldb : str
            Calibration database
        irf : str
            Instrument response function

        Returns
        obs : `~gammalib.GCTAObservation`
            CTA observation with response attached
        -------
        """
        # Create XML element
        xml = gammalib.GXmlElement()

        # Append parameter
        parameter = 'parameter name="Calibration" database="'+caldb+\
                    '" response="'+irf+'"'
        xml.append(gammalib.GXmlElement(parameter))

        # Create CTA response
        response = gammalib.GCTAResponseIrf()
        response.read(xml)

        # Attach response to observation
        obs.response(response)

        # Return observation
        return obs

 
    # Public methods
    def run(self):
        """
        Run the script

        Raises
        ------
        RuntimeError
            Invalid pointing definition file format
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Write header into logger
        self._log_header1(gammalib.TERSE,
                          'Creating observation definition XML file')

        # Load pointing definition file if it is not already set. Extract
        # the number of columns and pointings
        if self._pntdef.size() == 0:
            self._pntdef = gammalib.GCsv(self['inpnt'].filename(), ',')
        ncols = self._pntdef.ncols()
        npnt  = self._pntdef.nrows()-1

        # Raise an exception if there is no header information
        if self._pntdef.nrows() < 1:
            raise RuntimeError('No header found in pointing definition file.')

        # Clear observation container
        self._obs.clear()

        # Initialise observation identifier counter
        identifier = 1

        # Extract header columns from pointing definition file and put them
        # into a list
        header = []
        for col in range(ncols):
            header.append(self._pntdef[0,col])

        # Loop over all pointings
        for pnt in range(npnt):

            # Set pointing definition CSV file row index
            row = pnt + 1

            # Create empty CTA observation
            obs = gammalib.GCTAObservation()

            # Set observation name. If no observation name was given then
            # use "None".
            if 'name' in header:
                name = self._pntdef[row, header.index('name')]
            else:
                name = self['name'].string()
            obs.name(name)

            # Set observation identifier. If no observation identified was
            # given the use the internal counter.
            if 'id' in header:
                obsid = self._pntdef[row, header.index('id')]
            else:
                obsid = '%6.6d' % identifier
                identifier += 1
            obs.id(obsid)

            # Set pointing. Either use "ra" and "dec" or "lon" and "lat".
            # If none of these pairs are given then raise an exception.
            if 'ra' in header and 'dec' in header:
                ra     = float(self._pntdef[row, header.index('ra')])
                dec    = float(self._pntdef[row, header.index('dec')])
                pntdir = gammalib.GSkyDir()
                pntdir.radec_deg(ra,dec)
            elif 'lon' in header and 'lat' in header:
                lon    = float(self._pntdef[row, header.index('lon')])
                lat    = float(self._pntdef[row, header.index('lat')])
                pntdir = gammalib.GSkyDir()
                pntdir.lb_deg(lon,lat)
            else:
                raise RuntimeError('No (ra,dec) or (lon,lat) columns '
                                   'found in pointing definition file.')
            obs.pointing(gammalib.GCTAPointing(pntdir))

            # Set response function. If no "caldb" or "irf" information is
            # provided then use the user parameter values.
            if 'caldb' in header:
                caldb = self._pntdef[row, header.index('caldb')]
            else:
                caldb = self['caldb'].string()
            if 'irf' in header:
                irf = self._pntdef[row, header.index('irf')]
            else:
                irf = self['irf'].string()
            if caldb != '' and irf != '':
                obs = self._set_irf(obs, caldb, irf)

            # Set deadtime correction factor. If no information is provided
            # then use the user parameter value "deadc".
            if 'deadc' in header:
                deadc = float(self._pntdef[row, header.index('deadc')])
            else:
                deadc = self['deadc'].real()
            obs.deadc(deadc)

            # Set Good Time Interval. If no information is provided then use
            # the user parameter values "tmin" and "duration".
            if 'tmin' in header:
                self._tmin = float(self._pntdef[row, header.index('tmin')])
            if 'duration' in header:
                duration = float(self._pntdef[row, header.index('duration')])
            else:
                duration = self['duration'].real()
            tref       = gammalib.GTimeReference(self['mjdref'].real(),'s')
            tmin       = self._tmin
            tmax       = self._tmin + duration
            gti        = gammalib.GGti(tref)
            tstart     = gammalib.GTime(tmin, tref)
            tstop      = gammalib.GTime(tmax, tref)
            self._tmin = tmax
            gti.append(tstart, tstop)
            obs.ontime(gti.ontime())
            obs.livetime(gti.ontime()*deadc)

            # Set Energy Boundaries. If no "emin" or "emax" information is
            # provided then use the user parameter values in case they are
            # valid.
            has_emin = False
            has_emax = False
            if 'emin' in header:
                emin     = float(self._pntdef[row, header.index('emin')])
                has_emin = True
            else:
                if self['emin'].is_valid():
                    emin     = self['emin'].real()
                    has_emin = True
            if 'emax' in header:
                emax     = float(self._pntdef[row, header.index('emax')])
                has_emax = True
            else:
                if self['emax'].is_valid():
                    emax     = self['emax'].real()
                    has_emax = True
            has_ebounds = has_emin and has_emax
            if has_ebounds:
                ebounds = gammalib.GEbounds(gammalib.GEnergy(emin, 'TeV'),
                                            gammalib.GEnergy(emax, 'TeV'))

            # Set ROI. If no ROI radius is provided then use the user
            # parameters "rad".
            has_roi = False
            if 'rad' in header:
                rad     = float(self._pntdef[row, header.index('rad')])
                has_roi = True
            else:
                if self['rad'].is_valid():
                    rad     = self['rad'].real()
                    has_roi = True
            if has_roi:
                roi = gammalib.GCTARoi(gammalib.GCTAInstDir(pntdir), rad)

            # Create an empty event list
            event_list = gammalib.GCTAEventList()
            event_list.gti(gti)

            # If available, set the energy boundaries and the ROI
            if has_ebounds:
                event_list.ebounds(ebounds)
            if has_roi:
                event_list.roi(roi)

            # Attach event list to CTA observation
            obs.events(event_list)

            # Write observation into logger
            name  = obs.instrument()+' observation'
            value = 'Name="%s" ID="%s"' % (obs.name(), obs.id())
            self._log_value(gammalib.NORMAL, name, value)
            self._log_string(gammalib.EXPLICIT, str(obs)+'\n')

            # Append observation
            self._obs.append(obs)

        # Return
        return

    def save(self):
        """
        Save observation definition XML file.
        """
        # Write header and filename into logger
        self._log_header1(gammalib.TERSE, 'Save observation definition XML file')

        # Check if observation definition XML file is valid
        if self['outobs'].is_valid():

            # Get output filename in case it was not read ahead
            outobs = self['outobs'].filename()

            # Log filename
            self._log_value(gammalib.NORMAL, 'Observation XML file', outobs.url())

            # Save observation definition XML file
            self._obs.save(outobs)
        
        # Return
        return

    def obs(self):
        """
        Returns observation container
        -----
        Returns
        obs : `~gammalib.GObservations`
            Observation container
        """
        # Return container
        return self._obs

    def pntdef(self, csv):
        """
        Set pointing definition from a CSV table

        Parameters
        ----------
        csv : `~gammalib.GCsv`
            Comma-separated values table
        """
        # Set pointing definition
        self._pntdef = csv

        # Return
        return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csobsdef(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the ctfindvar tool.
#
# Copyright (C) 2018-2019 Simon Bonnefoy
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import ctools
from testing import test


# ======================= #
# Test class for ctfindvar tool #
# ======================= #
class Test(test):
    """
    Test class for ctfindvar tool

    This test class makes unit tests for the ctfindvar tool by using it from
    the command line and from Python.
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('ctfindvar')

        # Append tests
        self.append(self._test_cmd, 'Test ctfindvar on command line')
        self.append(self._test_python, 'Test ctfindvar from Python')

        # Return
        return

    # Test ctfindvar on command line
    def _test_cmd(self):
        """
        Test ctfindvar on the command line
        """
        # Set tool name
        ctfindvar = self._tool('ctfindvar')

        # Setup ctfindvar command
        cmd = ctfindvar+' inobs="'+self._events+'"'+\
                        ' outmap="ctfindvar_cmd1.fits"'+\
                        ' outmodel="ctfindvar_cmd1.xml"'+\
                        ' logfile="ctfindvar_cmd1.log" chatter=1'+\
                        ' tmin=NONE tmax=NONE tinterval=20'+\
                        ' emin=0.2 emax=1.1'+\
                        ' nxpix=200 nypix=200 binsz=0.05'+\
                        ' proj=CAR coordsys=CEL'+\
                        ' xref=83.63 yref=22.51'+\
                        ' xsrc=83.63 ysrc=22.51'+\
                        ' caldb=prod2 irf=South_0.5h'
        
        # Check if execution was successful
        self.test_value(self._execute(cmd), 0,
             'Check successful execution from command line')

        # Check result
        self._check_result_files('ctfindvar_cmd1.fits', 'ctfindvar_cmd1.xml', 0)

        # Setup ctfindvar command
        cmd = ctfindvar+' inobs="event_file_that_does_not_exist.fits"'+ \
                        ' outmap="ctfindvar_cmd2.fits"'+ \
                        ' outmodel="ctfindvar_cmd2.xml"'+ \
                        ' logfile="ctfindvar_cmd2.log" chatter=1'+ \
                        ' tmin=NONE tmax=NONE tinterval=20'+ \
                        ' emin=0.2 emax=1.1'+ \
                        ' nxpix=200 nypix=200 binsz=0.05'+ \
                        ' proj=CAR coordsys=CEL'+ \
                        ' xref=83.63 yref=22.51'+ \
                        ' xsrc=83.63 ysrc=22.51'+ \
                        ' caldb=prod2 irf=South_0.5h'
        
        # Check if execution was successful
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check ctfindvar --help
        self._check_help(ctfindvar)

        # Return
        return

    # Test ctfindvar from Python
    def _test_python(self):
        """
        Test ctfindvar from Python
        """
        # Allocate ctfindvar
        ctfindvar = ctools.ctfindvar()

        # Check that saving does not nothing
        ctfindvar['logfile']  = 'ctfindvar_py0.log'
        ctfindvar['outmap']   = 'ctfindvar_py0.fits'
        ctfindvar['outmodel'] = 'ctfindvar_py0.xml'
        ctfindvar.logFileOpen()
        ctfindvar.save()
        self.test_assert(not os.path.isfile('ctfindvar_py0.fits'),
             'Check that no output map has been created')
        self.test_assert(not os.path.isfile('ctfindvar_py0.xml'),
             'Check that no output model definition XML file has been created')

        # Check that clearing does not lead to an exception or segfault
        ctfindvar.clear()

        # Now set ctfindvar parameters
        ctfindvar['inobs']     = self._events
        ctfindvar['tmin']      = 'NONE'
        ctfindvar['tmax']      = 'NONE' 
        ctfindvar['tinterval'] = 20 
        ctfindvar['emin']      = 0.2 
        ctfindvar['emax']      = 1.1 
        ctfindvar['nxpix']     = 200 
        ctfindvar['nypix']     = 200 
        ctfindvar['binsz']     = 0.05 
        ctfindvar['coordsys']  = 'CEL'
        ctfindvar['proj']      = 'CAR'
        ctfindvar['xref']      = 83.63 
        ctfindvar['yref']      = 22.51
        ctfindvar['xsrc']      = 83.63
        ctfindvar['ysrc']      = 22.51
        ctfindvar['caldb']     = 'prod2'
        ctfindvar['irf']       = 'South_0.5h' 
        ctfindvar['outmap']    = 'ctfindvar_py1.fits'
        ctfindvar['outmodel']  = 'ctfindvar_py1.xml'
        ctfindvar['logfile']   = 'ctfindvar_py1.log'
        ctfindvar['chatter']   =  2

        # Execute ctfindvar tool
        ctfindvar.logFileOpen()
        ctfindvar.execute()

        # Check result
        self._check_result_files('ctfindvar_py1.fits', 'ctfindvar_py1.xml', 0)

        # Copy ctfindvar tool
        cpy_tool = ctfindvar.copy()

        # Execute copy of ctfindvar tool again with lower threshold and now
        # with output cube file
        cpy_tool['threshold'] = 1.0
        cpy_tool['outcube']   = 'ctfindvar_py2_cube.fits'
        cpy_tool['outmap']    = 'ctfindvar_py2.fits'
        cpy_tool['outmodel']  = 'ctfindvar_py2.xml'
        cpy_tool['logfile']   = 'ctfindvar_py2.log'
        cpy_tool['chatter']   = 3
        cpy_tool.logFileOpen()
        cpy_tool.execute()

        # Check result
        self._check_result_files('NONE', 'ctfindvar_py2.xml', 1)
        fits = gammalib.GFits('ctfindvar_py2_cube.fits')
        self.test_value(fits.size(), 1, 'Check for 1 extension in cube file')

        # Now run with an input model and time window, a smoothing kernel
        # and set the number of threads
        ctfindvar.clear()
        ctfindvar['inobs']         = self._events
        ctfindvar['inmodel']       = self._model
        ctfindvar['tmin']          = '2020-01-01T00:01:10'
        ctfindvar['tmax']          = '2020-01-01T00:03:40'
        ctfindvar['tinterval']     = 10
        ctfindvar['smooth_kernel'] = 'DISK'
        ctfindvar['smooth_rad']    = 0.05
        ctfindvar['emin']          = 0.2
        ctfindvar['emax']          = 1.1
        ctfindvar['nxpix']         = 200
        ctfindvar['nypix']         = 200
        ctfindvar['binsz']         = 0.05
        ctfindvar['coordsys']      = 'CEL'
        ctfindvar['proj']          = 'CAR'
        ctfindvar['xref']          = 83.63
        ctfindvar['yref']          = 22.51
        ctfindvar['caldb']         = 'prod2'
        ctfindvar['irf']           = 'South_0.5h'
        ctfindvar['outmap']        = 'ctfindvar_py3.fits'
        ctfindvar['outmodel']      = 'ctfindvar_py3.xml'
        ctfindvar['logfile']       = 'ctfindvar_py3.log'
        ctfindvar['chatter']       =  4
        ctfindvar['nthreads']      =  2

        # Execute ctfindvar tool
        ctfindvar.logFileOpen()
        ctfindvar.execute()

        # Check result
        self._check_result_files('ctfindvar_py3.fits', 'ctfindvar_py3.xml', 0)

        # Return
        return

    # Check ctfindvar result
    def _check_result_files(self, fitsfile, xmlfile, nmodels):
        """
        Check ctfindvar result

        Parameters
        ----------
        fitsfile : str
            FITS file name
        xmlfile : str
            Model definition XML file name
        nmodels : int
            Expected number of models
        """
        # Read variability FITS file
        if fitsfile is not 'NONE':
            fits = gammalib.GFits(fitsfile)
            self.test_value(fits.size(), 3, 'Check for 3 extensions in output file')

        # Read model definition XML file
        models = gammalib.GModels(xmlfile)
        self.test_value(models.size(), nmodels,
             'Check for %d models in model definition XML file' % nmodels)

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the ctmapcube tool.
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import ctools
from testing import test


# ============================= #
# Test class for ctmapcube tool #
# ============================= #
class Test(test):
    """
    Test class for ctmapcube tool

    This test class makes unit tests for the ctmapcube tool by using it from
    the command line and from Python.
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('ctmapcube')

        # Append tests
        self.append(self._test_cmd, 'Test ctmapcube on command line')
        self.append(self._test_python, 'Test ctmapcube from Python')

        # Return
        return

    # Test ctmapcube on command line
    def _test_cmd(self):
        """
        Test ctmapcube on the command line
        """
        # Set tool name
        ctmapcube = self._tool('ctmapcube')

        # Setup ctmapcube command
        cmd = ctmapcube+' inmodel="'+self._model+'"'+ \
                        ' outcube="ctmapcube_cmd1.fits"'+\
                        ' emin=1.0 emax=100.0 enumbins=5 ebinalg=LOG'+ \
                        ' nxpix=20 nypix=20 binsz=0.2 coordsys=CEL'+ \
                        ' xref=83.63 yref=22.01 proj=CAR'+ \
                        ' logfile="ctmapcube_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check map cube
        self._check_result_file('ctmapcube_cmd1.fits')

        # Setup ctmapcube command
        cmd = ctmapcube+' inmodel="events_that_do_not_exist.fits"'+ \
                        ' outcube="ctmapcube_cmd2.fits"'+\
                        ' emin=1.0 emax=100.0 enumbins=5 ebinalg=LOG'+ \
                        ' nxpix=20 nypix=20 binsz=0.2 coordsys=CEL'+ \
                        ' xref=83.63 yref=22.01 proj=CAR'+ \
                        ' logfile="ctmapcube_cmd2.log" debug=yes '

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check ctmapcube --help
        self._check_help(ctmapcube)

        # Return
        return

    # Test ctmapcube from Python
    def _test_python(self):
        """
        Test ctmapcube from Python
        """
        # Allocate ctmapcube
        mapcube = ctools.ctmapcube()

        # Check that empty ctmapcube tool holds an map cube that has no
        # energy bins
        self._check_cube(mapcube.mapcube(), nmaps=0, npixels=0)

        # Check that saving does not nothing
        mapcube['outcube'] = 'ctmapcube_py0.fits'
        mapcube['logfile'] = 'ctmapcube_py0.log'
        mapcube.logFileOpen()
        mapcube.save()
        self.test_assert(not os.path.isfile('ctmapcube_py0.fits'),
             'Check that no map cube has been created')

        # Check that clearing does not lead to an exception or segfault
        mapcube.clear()

        # Now set ctmapcube parameters
        mapcube['inmodel']  = self._model
        mapcube['ebinalg']  = 'LOG'
        mapcube['emin']     = 1.0
        mapcube['emax']     = 100.0
        mapcube['enumbins'] = 5
        mapcube['nxpix']    = 20
        mapcube['nypix']    = 20
        mapcube['binsz']    = 0.2
        mapcube['coordsys'] = 'CEL'
        mapcube['proj']     = 'CAR'
        mapcube['xref']     = 83.63
        mapcube['yref']     = 22.01
        mapcube['outcube']  = 'ctmapcube_py1.fits'
        mapcube['logfile']  = 'ctmapcube_py1.log'
        mapcube['chatter']  = 2

        # Run ctmapcube tool
        mapcube.logFileOpen()   # Make sure we get a log file
        mapcube.run()

        # Save map cube
        mapcube.save()

        # Check map cube
        self._check_result_file('ctmapcube_py1.fits')

        # Copy ctmapcube tool
        cpy_mapcube = mapcube.copy()

        # Check map cube of ctmapcube copy
        self._check_cube(cpy_mapcube.mapcube())

        # Execute copy of ctmapcube tool again, now with a higher chatter
        # level than before
        cpy_mapcube['outcube'] = 'ctmapcube_py2.fits'
        cpy_mapcube['logfile'] = 'ctmapcube_py2.log'
        cpy_mapcube['publish'] = True
        cpy_mapcube['chatter'] = 3
        cpy_mapcube.logFileOpen()  # Needed to get a new log file
        cpy_mapcube.execute()

        # Check result file
        self._check_result_file('ctmapcube_py2.fits')

        # Now clear copy of ctmapcube tool
        cpy_mapcube.clear()

        # Check that the cleared copy has also cleared the map cube
        self._check_cube(cpy_mapcube.mapcube(), nmaps=0, npixels=0)

        # Allocate ctmapcube tools, set models and do now a linear
        # binning; also do not use Gaussian point sources
        mapcube = ctools.ctmapcube()
        mapcube.models(gammalib.GModels(self._model))
        mapcube['ebinalg']  = 'LIN'
        mapcube['emin']     = 1.0
        mapcube['emax']     = 100.0
        mapcube['enumbins'] = 5
        mapcube['nxpix']    = 20
        mapcube['nypix']    = 20
        mapcube['binsz']    = 0.2
        mapcube['coordsys'] = 'CEL'
        mapcube['proj']     = 'CAR'
        mapcube['xref']     = 83.63
        mapcube['yref']     = 22.01
        mapcube['ptsrcsig'] = 0.0
        mapcube['outcube']  = 'ctmapcube_py3.fits'
        mapcube['logfile']  = 'ctmapcube_py3.log'
        mapcube['chatter']  = 4

        # Execute mapcube
        mapcube.logFileOpen()  # Needed to get a new log file
        mapcube.execute()

        # Check result file
        self._check_result_file('ctmapcube_py3.fits')

        # Publish with a name
        mapcube.publish('My map cube')

        # Return
        return

    # Check result file
    def _check_result_file(self, filename):
        """
        Check content of map cube

        Parameters
        ----------
        filename : str
            Map cube file name
        """
        # Load map cube
        cube = gammalib.GModelSpatialDiffuseCube(filename)

        # Check map cube
        self._check_cube(cube)

        # Return
        return

    # Check map cube
    def _check_cube(self, cube, nmaps=6, npixels=400):
        """
        Check map cube

        Parameters
        ----------
        cube : `~gammalib.GModelSpatialDiffuseCube`
            Map cube
        nmaps : int, optional
            Number of maps
        npixels : int, optional
            Number of pixels
        """
        # Get energies (this forces loading in case the map cube is not
        # loaded)
        self.test_value(len(cube.energies()), nmaps, 'Check number of energies')
        
        # Check dimensions
        self.test_value(cube.maps(), nmaps, 'Check number of maps')
        self.test_value(cube.pixels(), npixels, 'Check number of pixels')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Display pointings in observation definition XML file
#
# Copyright (C) 2017 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import cscripts
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()


# =============================== #
# Extract pointings from XML file #
# =============================== #
def get_pointings(filename):
    """
    Extract pointings from XML file

    Parameters
    ----------
    filename : str
        File name of observation definition XML file

    Returns
    -------
    pnt : list of dict
        Pointings
    """
    # Initialise pointings
    pnt = []

    # Open XML file
    xml = gammalib.GXml(filename)

    # Get observation list
    obs = xml.element('observation_list')

    # Get number of observations
    nobs = obs.elements('observation')

    # Loop over observations
    for i in range(nobs):

        # Get observation
        run = obs.element('observation', i)

        # Get pointing parameter
        npars  = run.elements('parameter')
        ra     = None
        dec    = None
        south  = False
        evfile = None
        for k in range(npars):
            par = run.element('parameter', k)
            if par.attribute('name') == 'Pointing':
                ra  = float(par.attribute('ra'))
                dec = float(par.attribute('dec'))
            elif par.attribute('name') == 'Calibration':
                if 'South' in par.attribute('response'):
                    south = True
                else:
                    south = False
            elif par.attribute('name') == 'EventList':
                evfile = par.attribute('file')

        # If no pointing was found then load observation
        if ra == None and evfile != None:
            cta = gammalib.GCTAObservation(gammalib.GFilename(evfile))
            ra  = cta.pointing().dir().ra_deg()
            dec = cta.pointing().dir().dec_deg()

        # Add valid pointing
        if ra != None:
            p   = gammalib.GSkyDir()
            p.radec_deg(ra, dec)
            entry = {'l': p.l_deg(), 'b': p.b_deg(), 'ra': ra, 'dec': dec, 'south': south}
            pnt.append(entry)

    # Return pointings
    return pnt


# ================ #
# Plot information #
# ================ #
def plot_pointings(pnt, plotfile):
    """
    Plot information

    Parameters
    ----------
    pnt : list of dict
        Pointings
    plotfile : str
        Plot filename
    """
    # Create figure
    plt.figure()

    # Setup figure
    ax = plt.gca()
    ax.cla()
    ax.set_xlim((180, -180))
    ax.set_ylim((-90, 90))

    # Loop over pointings
    for p in pnt:

        # Get longitude
        l = p['l']
        if l > 180.0:
            l = l - 360.0

        # Set color
        if p['south']:
            color='r'
        else:
            color='b'

        # Set circle
        circle = plt.Circle((l, p['b']), 3.0, color=color, fill=False)

        # Add circle
        ax.add_artist(circle)

    # Plot title and labels
    plt.xlabel('Galactic longitude (deg)')
    plt.ylabel('Galactic latitude (deg)')
    plt.title('Pointings')

    # Show plots or save it into file
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ============== #
# Show pointings #
# ============== #
def show_pointings():
    """
    Show pointings
    """
    # Set usage string
    usage = 'show_pointings.py [-p plotfile] file'

    # Set default options
    options = [{'option': '-p', 'value': ''}]

    # Get arguments and options from command line arguments
    args, options = cscripts.ioutils.get_args_options(options, usage)

    # Extract script parameters from options
    plotfile = options[0]['value']

    # Get pointings
    pnt = get_pointings(args[0])

    # Plot pointings
    plot_pointings(pnt, plotfile)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Show pointings
    show_pointings()
